#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import json
import os

from distutils import util as dist_utils

import yaml

import directord

from directord import utils


class Mixin:
    """Mixin class."""

    def __init__(self, args):
        """Initialize the Directord mixin.

        Sets up the mixin object.

        :param args: Arguments parsed by argparse.
        :type args: Object
        """

        self.args = args

    def format_action(
        self,
        verb,
        execute,
        arg_vars=None,
        targets=None,
        ignore_cache=False,
        restrict=None,
        parent_id=None,
        parent_sha3_224=None,
        return_raw=False,
        parent_async=False,
    ):
        """Return a JSON encode object for task execution.

        While formatting the message, the method will treat each verb as a
        case and parse the underlying sub-command, formatting the information
        into a dictionary.

        :param verb: Action to parse.
        :type verb: String
        :param execute: Execution string to parse.
        :type execute: String
        :param targets: Target argents to send job to.
        :type targets: List
        :param arg_vars: Argument dictionary, used to set arguments in
                         dictionary format instead of string format.
        :type arg_vars: Dictionary
        :param ignore_cache: Instruct the entire execution to
                             ignore client caching.
        :type ignore_cache: Boolean
        :param restrict: Restrict job execution based on a provided task
                         SHA3_224.
        :type restrict: List
        :param parent_id: Set the parent UUID for execution jobs.
        :type parent_id: String
        :param parent_sha3_224: Set the parent sha3_224 for execution jobs.
        :type parent_id: String
        :param return_raw: Enable a raw return from the server.
        :type return_raw: Boolean
        :param parent_async: Enable a parent job to run asynchronously.
        :type parent_async: Boolean
        :returns: String
        """

        data = dict(verb=verb)
        component_kwargs = dict(
            exec_array=execute, data=data, arg_vars=arg_vars
        )

        success, transfer, component = directord.component_import(
            component=verb.lower(),
            job_id=parent_id,
        )
        if not success:
            raise SystemExit(component)

        setattr(component, "verb", verb)
        data.update(component.server(**component_kwargs))

        data["timeout"] = getattr(component.known_args, "timeout", 600)
        data["run_once"] = getattr(component.known_args, "run_once", False)
        data["job_sha3_224"] = utils.object_sha3_224(obj=data)
        data["return_raw"] = return_raw
        data["skip_cache"] = ignore_cache or getattr(
            component.known_args, "skip_cache", False
        )

        if targets:
            data["targets"] = targets

        if parent_async:
            data["parent_async"] = parent_async

        if parent_id:
            data["parent_id"] = parent_id

        if parent_sha3_224:
            data["parent_sha3_224"] = parent_sha3_224

        if restrict:
            data["restrict"] = restrict

        if transfer:
            job = {
                "jobs": [
                    {"WORKDIR": "/etc/directord/components"},
                    {
                        "ADD": "--skip-cache {} {}".format(
                            transfer, "/etc/directord/components/"
                        )
                    },
                ]
            }
            self.exec_orchestrations(
                orchestrations=[job],
                defined_targets=data.get("targets"),
                return_raw=True,
            )

        return json.dumps(data)

    def exec_orchestrations(
        self,
        orchestrations,
        defined_targets=None,
        restrict=None,
        ignore_cache=False,
        return_raw=False,
    ):
        """Execute orchestration jobs.

        Iterates over a list of orchestartion blobs, fingerprints the jobs,
        and then runs them.

        :param orchestrations: List of Dictionaries which are run as
                               orchestartion.
        :type orchestrations: List
        :param defined_targets: List of targets to limit a given execution to.
                                This target list provides an override for
                                targets found within a given orchestation.
        :type defined_targets: List
        :param restrict: Restrict a given orchestration job to a set of
                         SHA3_224 job fingerprints.
        :type restrict: Array
        :param ignore_cache: Instruct the orchestartion job to ignore cached
                             executions.
        :type ignore_cache: Boolean
        :param return_raw: Enable a raw return from the server.
        :type return_raw: Boolean
        :returns: List
        """

        job_to_run = list()
        for orchestrate in orchestrations:
            parent_sha3_224 = utils.object_sha3_224(obj=orchestrate)
            parent_id = utils.get_uuid()
            targets = defined_targets or orchestrate.get("targets", list())

            force_async = getattr(self.args, "force_async", False)
            if force_async:
                parent_async = force_async
            else:
                try:
                    parent_async = bool(
                        dist_utils.strtobool(orchestrate.get("async", "False"))
                    )
                except (ValueError, AttributeError):
                    parent_async = bool(orchestrate.get("async", False))

            for job in orchestrate["jobs"]:
                arg_vars = job.pop("vars", None)
                key, value = next(iter(job.items()))
                job_to_run.append(
                    dict(
                        verb=key,
                        execute=[value],
                        arg_vars=arg_vars,
                        targets=targets,
                        restrict=restrict,
                        ignore_cache=ignore_cache,
                        parent_id=parent_id,
                        parent_sha3_224=parent_sha3_224,
                        return_raw=return_raw,
                        parent_async=parent_async,
                    )
                )

        return_data = list()
        if getattr(self.args, "finger_print", False):
            count = 0
            for job in job_to_run:
                tabulated_data = list()
                formatted_job = self.format_action(**job)
                item = json.loads(formatted_job)
                exec_str = " ".join(job["execute"])
                if len(exec_str) >= 30:
                    exec_str = "{execute}...".format(execute=exec_str[:27])
                tabulated_data.extend(
                    [
                        count,
                        job["parent_sha3_224"],
                        item["verb"],
                        exec_str,
                        item["job_sha3_224"],
                    ]
                )
                return_data.append(tabulated_data)
                count += 1
            utils.print_tabulated_data(
                data=return_data,
                headers=["count", "parent_sha", "verb", "exec", "job_sha"],
            )
            return []
        else:
            for job in job_to_run:
                formatted_job = self.format_action(**job)
                return_data.append(
                    directord.send_data(
                        socket_path=self.args.socket_path, data=formatted_job
                    )
                )

        return return_data

    def run_orchestration(self):
        """Run orchestration jobs.

        When orchestration jobs are executed the files are organized and
        then indexed. Once indexed, the jobs are sent to the server. send
        returns are captured and returned on method exit.

        :returns: List
        """

        return_data = list()
        for orchestrate_file in self.args.orchestrate_files:
            orchestrate_file = os.path.abspath(
                os.path.expanduser(orchestrate_file)
            )
            if not os.path.exists(orchestrate_file):
                raise FileNotFoundError(
                    "The [ {} ] file was not found.".format(orchestrate_file)
                )
            else:
                with open(orchestrate_file) as f:
                    orchestrations = yaml.safe_load(f)

                if self.args.target:
                    defined_targets = list(set(self.args.target))
                else:
                    defined_targets = list()

                return_data.extend(
                    self.exec_orchestrations(
                        orchestrations=orchestrations,
                        defined_targets=defined_targets,
                        restrict=self.args.restrict,
                        ignore_cache=self.args.ignore_cache,
                        return_raw=getattr(
                            self.args,
                            "poll",
                            False,
                        )
                        or getattr(self.args, "stream", False),
                    )
                )
        else:
            return return_data

    def run_exec(self):
        """Execute an exec job.

        Jobs are parsed and then sent to the server for processing. All return
        items are captured in an array which is returned on method exit.

        :returns: List
        """

        format_kwargs = dict(
            verb=self.args.verb,
            execute=self.args.exec,
            parent_async=getattr(self.args, "force_async", False),
            return_raw=getattr(self.args, "poll", False)
            or getattr(self.args, "stream", False),
        )
        if self.args.target:
            format_kwargs["targets"] = list(set(self.args.target))

        return [
            directord.send_data(
                socket_path=self.args.socket_path,
                data=self.format_action(**format_kwargs),
            )
        ]

    def return_tabulated_info(self, data):
        """Return a list of data that will be tabulated.

        :param data: Information to generally parse and return
        :type data: Dictionary
        :returns: List
        """

        data_id = data.pop("id", None) or self.args.job_info
        if data_id:
            tabulated_data = [["ID", data_id]]
        else:
            tabulated_data = list()

        for key, value in data.items():
            if value is None:
                continue

            if key.startswith("_"):
                continue

            if key == "PROCESSING":
                if value == b"\026":
                    value = "True"
                else:
                    value = "False"

            if isinstance(value, list):
                value = "\n".join(value)
            elif isinstance(value, dict):
                value = "\n".join(
                    ["{} = {}".format(k, v) for k, v in value.items() if v]
                )

            tabulated_data.append([key.upper(), value])
        else:
            return tabulated_data

    def return_tabulated_data(self, data, restrict_headings):
        """Return tabulated data displaying a limited set of information.

        :param data: Information to generally parse and return
        :type data: Dictionary
        :param restrict_headings: List of headings in string format to return
        :type restrict_headings: List
        :returns: List
        """

        def _computed_totals(item, value_heading, value):
            if item not in seen_computed_key:
                if isinstance(value, bool):
                    bool_heading = "{}_{}".format(value_heading, value).upper()
                    if bool_heading in computed_values:
                        computed_values[bool_heading] += 1
                    else:
                        computed_values[bool_heading] = 1
                elif isinstance(value, (float, int)):
                    if value_heading in computed_values:
                        computed_values[value_heading] += value
                    else:
                        computed_values[value_heading] = value

        tabulated_data = list()
        computed_values = dict()
        seen_computed_key = list()
        found_headings = ["ID"]
        original_data = list(dict(data).items())
        result_filter = getattr(self.args, "filter", None)
        for key, value in original_data:
            arranged_data = [key]
            include = result_filter is None
            for item in restrict_headings:
                if item not in found_headings:
                    found_headings.append(item)
                if item.upper() not in value and item.lower() not in value:
                    arranged_data.append(0)
                else:
                    try:
                        report_item = value[item.upper()]
                    except KeyError:
                        report_item = value[item.lower()]

                    if item.upper() == "PROCESSING":
                        if report_item.encode() == b"\026":
                            report_item = "True"
                            if result_filter == "processing":
                                include = True
                        else:
                            report_item = "False"
                    elif (
                        isinstance(report_item, list) and len(report_item) > 0
                    ):
                        if (
                            result_filter == "success"
                            and item.upper() == "SUCCESS"
                        ):
                            include = True
                        if (
                            result_filter == "failed"
                            and item.upper() == "FAILED"
                        ):
                            include = True

                    if not report_item:
                        arranged_data.append(0)
                    else:
                        if report_item and isinstance(report_item, list):
                            arranged_data.append(len(report_item))
                        elif isinstance(report_item, float):
                            arranged_data.append("{:.2f}".format(report_item))
                        else:
                            arranged_data.append(report_item)
                        _computed_totals(
                            item=key, value_heading=item, value=report_item
                        )

            if include:
                seen_computed_key.append(key)
                tabulated_data.append(arranged_data)

        return tabulated_data, found_headings, computed_values



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
import json
import os
import sys

import jinja2
from jinja2 import StrictUndefined

import yaml

import directord

from directord import bootstrap
from directord import client
from directord import meta
from directord import mixin
from directord import server
from directord import user
from directord import utils


def _args(exec_args=None):
    """Setup client arguments."""

    parser = argparse.ArgumentParser(
        description="Deployment Framework Next.",
        prog="Directord",
        formatter_class=lambda prog: argparse.HelpFormatter(
            prog, max_help_position=32, width=128
        ),
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=directord.__version__),
    )
    parser.add_argument(
        "--config-file",
        help="File path for client configuration. Default: %(default)s",
        metavar="STRING",
        default=os.getenv("DIRECTORD_CONFIG_FILE", None),
        type=argparse.FileType(mode="r"),
    )
    parser.add_argument(
        "--datastore",
        help=(
            "Connect to an external datastore for job and worker tracking. The"
            " connection string is RFC-1738 compatible >"
            " driver://username:password@host:port/database. If undefined the"
            " datastore uses an internal manager object. Driver"
            " supports ['redis', 'file']. When using Redis, two"
            " keyspaces will be used and are incremented by 1 using the"
            " database value (which has a default of 0). If the datastore"
            " option is set to 'memory', the server will spawn a manager"
            " thread to facilitate the document store."
        ),
        metavar="STRING",
        default=os.getenv(
            "DIRECTORD_DATASTORE", "file:///var/cache/directord"
        ),
        type=str,
    )
    parser.add_argument(
        "--driver",
        help="Messaging driver used for workload transport.",
        default=os.getenv("DIRECTORD_DRIVER", meta.__driver_default__),
        choices=meta.__driver_options__,
        type=str,
    )
    auth_group = parser.add_mutually_exclusive_group()
    auth_group.add_argument(
        "--shared-key",
        help="Shared key used for server client authentication.",
        metavar="STRING",
        default=os.getenv("DIRECTORD_SHARED_KEY", None),
    )
    auth_group.add_argument(
        "--curve-encryption",
        action="store_true",
        help=(
            "Server and client will connect using Curve authentication"
            " and encryption. Enabling this option assumes keys have been"
            " generated. see `manage --generate-keys` for more."
        ),
    )
    parser.add_argument(
        "--debug",
        help="Enable debug mode. Default: %(default)s",
        action="store_true",
    )
    parser.add_argument(
        "--job-port",
        help="Job port to bind. Default: %(default)s",
        metavar="INT",
        default=int(os.getenv("DIRECTORD_MSG_PORT", 5555)),
        type=int,
    )
    parser.add_argument(
        "--backend-port",
        help="Backend port to bind. Default: %(default)s",
        metavar="INT",
        default=int(os.getenv("DIRECTORD_BACKEND_PORT", 5556)),
        type=int,
    )
    parser.add_argument(
        "--heartbeat-interval",
        help="heartbeat interval in seconds. Default: %(default)s",
        metavar="INT",
        default=int(os.getenv("DIRECTORD_HEARTBEAT_INTERVAL", 60)),
        type=int,
    )
    parser.add_argument(
        "--socket-path",
        help=(
            "Server file socket path for user interactions."
            " Default: %(default)s"
        ),
        metavar="STRING",
        default=str(
            os.getenv("DIRECTORD_SOCKET_PATH", "/var/run/directord.sock")
        ),
        type=str,
    )
    parser.add_argument(
        "--socket-group",
        help=(
            "Server file socket group ownership for user interactions."
            " Default: %(default)s"
        ),
        metavar="STRING",
        default=str(os.getenv("DIRECTORD_SOCKET_GROUP", 0)),
    )
    parser.add_argument(
        "--cache-path",
        help=("Client cache path. Default: %(default)s"),
        metavar="STRING",
        default=str(os.getenv("DIRECTORD_CACHE_PATH", "/var/cache/directord")),
        type=str,
    )
    subparsers = parser.add_subparsers(
        help="Mode sub-command help", dest="mode"
    )
    parser_orchestrate = subparsers.add_parser(
        "orchestrate", help="Orchestration mode help"
    )
    parser_orchestrate.add_argument(
        "--restrict",
        help="Restrict orchestration to a set of Task SHA3_224(s).",
        metavar="STRING",
        nargs="+",
    )
    parser_orchestrate.add_argument(
        "--target",
        help="Worker target(s) to run a particular job against.",
        metavar="STRING",
        nargs="+",
    )
    parser_orchestrate.add_argument(
        "--ignore-cache",
        help=(
            "Instruct the orchestration engine to ignore all"
            " cache for the entirety of the run."
        ),
        action="store_true",
    )
    parser_orchestrate.add_argument(
        "--force-async",
        help=(
            "Instruct the orchestration engine to run all orchestrations"
            " asynchronously."
        ),
        action="store_true",
    )
    orchestrate_group = parser_orchestrate.add_mutually_exclusive_group(
        required=False
    )
    orchestrate_group.add_argument(
        "--poll",
        help="Block on client return for the completion of executed jobs.",
        action="store_true",
    )
    orchestrate_group.add_argument(
        "--finger-print",
        help="Finger print a set of orchestrations.",
        action="store_true",
    )
    orchestrate_group.add_argument(
        "--stream",
        help="Stream the STDOUT|STDERR for tasks.",
        action="store_true",
    )
    parser_orchestrate.add_argument(
        "--check",
        help=(
            "When polling is enabled this option can be used to check the"
            " outcome of a given task. If the task fails the client will"
            " fail too."
        ),
        action="store_true",
    )
    parser_orchestrate.add_argument(
        "orchestrate_files",
        help="YAML files to use for orchestration.",
        metavar="STRING",
        nargs="+",
    )
    parser_exec = subparsers.add_parser("exec", help="Execution mode help")
    parser_exec.add_argument(
        "--verb",
        help="Module Invocation for exec.",
        metavar="STRING",
        required=True,
    )
    parser_exec.add_argument(
        "--target",
        help="Worker target(s) to run a particular job against.",
        metavar="[STRING]",
        nargs="+",
    )
    parser_exec.add_argument(
        "exec",
        help="Freeform command. Use quotes for complex commands.",
        metavar="STRING",
        nargs="+",
    )
    parser_exec.add_argument(
        "--poll",
        help="Block on client return for the completion of executed jobs.",
        action="store_true",
    )
    parser_exec.add_argument(
        "--stream",
        help="Stream the STDOUT|STDERR for tasks.",
        action="store_true",
    )
    parser_exec.add_argument(
        "--check",
        help=(
            "When polling is enabled this option can be used to check the"
            " outcome of a given task. If the task fails the client will"
            " fail too."
        ),
        action="store_true",
    )
    parser_exec.add_argument(
        "--force-async",
        help=("Instruct the execution engine to run asynchronously."),
        action="store_true",
    )
    parser_server = subparsers.add_parser("server", help="Server mode help")
    parser_server.add_argument(
        "--bind-address",
        help="IP Address to bind a Directord Server. Default: %(default)s",
        metavar="STRING",
        default=os.getenv("DIRECTORD_BIND_ADDRESS", "*"),
    )
    parser_server.add_argument(
        "--run-ui",
        help="Enable the Directord UI. Default: %(default)s",
        action="store_true",
    )
    parser_server.add_argument(
        "--ui-port",
        help="UI server bind port. Default: %(default)s",
        metavar="INT",
        default=int(os.getenv("DIRECTORD_UI_PORT", 9000)),
        type=int,
    )
    parser_client = subparsers.add_parser("client", help="Client mode help")
    parser_client.add_argument(
        "--server-address",
        help=(
            "Domain or IP address of the Directord server."
            " Default: %(default)s"
        ),
        metavar="STRING",
        default=os.getenv("DIRECTORD_SERVER_ADDRESS", "localhost"),
    )
    parser_manage = subparsers.add_parser(
        "manage", help="Server management mode help"
    )
    filter_group = parser_manage.add_mutually_exclusive_group(required=False)
    filter_group.add_argument(
        "--filter",
        choices=["success", "failed", "processing"],
        help="List filtered jobs.",
    )
    manage_group = parser_manage.add_mutually_exclusive_group(required=True)
    manage_group.add_argument(
        "--list-jobs", action="store_true", help="List all known jobs."
    )
    manage_group.add_argument(
        "--list-nodes", action="store_true", help="List all available nodes."
    )
    manage_group.add_argument(
        "--purge-jobs",
        action="store_true",
        help="Purge all jobs from the server.",
    )
    manage_group.add_argument(
        "--purge-nodes",
        action="store_true",
        help="Purge all nodes from the server.",
    )
    manage_group.add_argument(
        "--job-info",
        help="Pull information on a specific job ID.",
        metavar="STRING",
    )
    manage_group.add_argument(
        "--export-jobs",
        help="Exports all job records as YAML and dumps them to a file.",
        metavar="STRING",
    )
    manage_group.add_argument(
        "--export-nodes",
        help="Exports all node records as YAML and dumps them to a file.",
        metavar="STRING",
    )
    manage_group.add_argument(
        "--generate-keys",
        action="store_true",
        help="Generate encryption keys for Curve authentication.",
    )
    manage_group.add_argument(
        "--dump-cache",
        action="store_true",
        help="Dump the local cache to stdout.",
    )
    manage_group.add_argument(
        "--analyze-parent",
        help="Analyze a given parent ID.",
        metavar="STRING",
    )
    manage_group.add_argument(
        "--analyze-job",
        help="Analyze a given job ID.",
        metavar="STRING",
    )
    parser_bootstrap = subparsers.add_parser(
        "bootstrap",
        help=(
            "Bootstrap a directord cluster. This uses SSH to connect to remote"
            " machines and setup Directord. Once Directord is setup, SSH is no"
            " longer required."
        ),
    )
    parser_bootstrap.add_argument(
        "--catalog",
        help="File path for SSH catalog.",
        metavar="STRING",
        action="append",
        required=True,
        type=argparse.FileType(mode="r"),
    )
    parser_bootstrap.add_argument(
        "--key-file",
        help=(
            "SSH Key file to use when connecting to targets."
            " Default: %(default)s"
        ),
        metavar="STRING",
        default=os.getenv("DIRECTORD_BOOTSTRAP_SSH_KEY_FILE"),
    )
    parser_bootstrap.add_argument(
        "--threads",
        help="Client bootstrap threads. Default: %(default)s",
        metavar="INT",
        default=int(os.getenv("DIRECTORD_BOOTSTRAP_THREADS", 10)),
        type=int,
    )
    if exec_args:
        args = parser.parse_args(args=exec_args)
    else:
        args = parser.parse_args()
    # Check for configuration file and load it if found.
    if args.config_file:
        config_data = yaml.safe_load(args.config_file)
        if config_data:
            for key, value in config_data.items():
                if isinstance(value, list) and key in args.__dict__:
                    args.__dict__[key].extend(value)
                else:
                    args.__dict__[key] = value

    return args, parser


class SystemdInstall:
    """Simple system service unit creation class."""

    def __init__(self, group="root", force=False):
        """Class to create systemd service units.

        This class is used with the directord-server-systemd and
        directord-client-systemd entrypoints.

        :param group: Name of the group used within the systemd service units.
        :type group: String
        :param force: Force install systemd service units.
        :type force: Boolean
        """

        self.config_path = "/etc/directord"
        self.socket_group = group
        self.force = force

    def path_setup(self):
        """Create the configuration path and basic configuration file."""

        os.makedirs(self.config_path, exist_ok=True)
        if not os.path.exists("/etc/directord/config.yaml"):
            with open("/etc/directord/config.yaml", "w") as f:
                f.write("---\ndebug: false\n")
            print("[+] Created empty configuration file")

    def writer(self, service_file):
        """Write a given systemd service unit file.

        :param service_file: Name of the embedded service file to
                             interact with.
        :type service_file: String
        """

        path = os.path.abspath(os.path.dirname(sys.argv[0]))
        self.path_setup()
        base = os.path.dirname(directord.__file__)
        service_file_path = "/etc/systemd/system/{}".format(service_file)
        if os.path.exists(service_file_path) and not self.force:
            print(
                "[-] Service file was not created because it already exists."
            )
            return

        blueprintLoader = jinja2.FileSystemLoader(
            searchpath=os.path.join(base, "templates")
        )
        blueprintEnv = jinja2.Environment(
            loader=blueprintLoader,
            keep_trailing_newline=True,
            undefined=StrictUndefined,
        )
        blueprint = blueprintEnv.get_template("{}.j2".format(service_file))
        blueprint_args = {
            "directord_binary": os.path.join(path, "directord"),
            "directord_group": self.socket_group,
        }
        outputText = blueprint.render(**blueprint_args)
        with open(service_file_path, "w") as f:
            f.write(outputText)

        print("[+] Installed {} service unit file".format(service_file))
        print(
            "[?] Run `systemctl daemon-reload` for unit file to take effect."
        )

    def server(self):
        """Run the server systemd service unit file creation process."""

        self.writer(service_file="directord-server.service")

    def client(self):
        """Run the client systemd service unit file creation process."""

        self.writer(service_file="directord-client.service")


def _systemd_loader():
    parser = argparse.ArgumentParser(
        description="Systemd install args.",
        prog="Directord-service",
        formatter_class=lambda prog: argparse.HelpFormatter(
            prog, max_help_position=32, width=128
        ),
    )
    parser.add_argument(
        "--group",
        help="Server group. Default: %(default)s.",
        metavar="STRING",
        default="root",
        type=str,
    )
    parser.add_argument(
        "--force",
        help="Force install systemd service unit file.",
        action="store_true",
    )
    args = parser.parse_args()
    return SystemdInstall(group=args.group, force=args.force)


def _systemd_server():
    """Execute the systemd server unit file creation process."""

    _systemd = _systemd_loader()
    _systemd.server()


def _systemd_client():
    """Execute the systemd client unit file creation process."""

    _systemd = _systemd_loader()
    _systemd.client()


def main():
    """Execute the main application.

    * Server|Client operations run within the foreground of the application.

    * Exec will submit a job to be run across the cluster.

    * Manage jobs will return information about the cluster and all executed
      jobs. This data will be returned in table format for easy consumptions.
    """

    args, parser = _args()
    _mixin = mixin.Mixin(args=args)

    if args.mode == "server":
        server.Server(args=args).worker_run()
    elif args.mode == "client":
        client.Client(args=args).worker_run()
    elif args.mode in ["exec", "orchestrate"]:
        if args.mode == "exec":
            return_data = _mixin.run_exec()
        else:
            return_data = _mixin.run_orchestration()

        job_items = [i.decode() for i in return_data if i]

        if args.poll or args.stream:
            failed = list()
            manage = user.Manage(args=args)
            for item in job_items:
                state, status, stdout, stderr, info = manage.poll_job(
                    job_id=item
                )
                if args.check and state is False:
                    failed.append(item)

                if args.stream:
                    for node in sorted(
                        set(
                            i for v in [stdout, stderr, info] for i in v.keys()
                        )
                    ):
                        for k, n, v in [
                            (node, name, d[node])
                            for name, d in [
                                ("STDOUT", stdout),
                                ("STDERR", stderr),
                                ("INFO", info),
                            ]
                            if node in d
                        ]:
                            print("{} -- {} {}".format(k, n, v))

                print(status)

            if any(failed):
                print("FAILED JOBS")
                for item in failed:
                    print(item)
                raise SystemExit(1)
        else:
            for item in job_items:
                print(item)

    elif args.mode == "manage":
        manage_exec = user.Manage(args=args)
        data = manage_exec.run()
        if args.generate_keys:
            print(
                "Keys generated. Synchronize the server and client public"
                " keys to client nodes to enable Curve encryption."
            )
            return

        try:
            data = json.loads(data)
        except Exception as e:
            print("No valid data found: {}".format(str(e)))
            return
        else:
            if not data:
                raise SystemExit("No data found")

        if args.export_jobs or args.export_nodes:
            export_file = utils.dump_yaml(
                file_path=(args.export_jobs or args.export_nodes),
                data=dict(data),
            )
            print("Exported data to [ {} ]".format(export_file))
            return

        computed_values = dict()
        headings = ["KEY", "VALUE"]
        tabulated_data = None
        if data and isinstance(data, dict):
            tabulated_data = _mixin.return_tabulated_info(data=data)
        elif data and isinstance(data, list):
            if args.job_info:
                item = dict(data).get(args.job_info)
                if not item:
                    print(
                        "Job information for ID:{} was not found".format(
                            args.job_info
                        )
                    )
                    return

                tabulated_data = _mixin.return_tabulated_info(data=item)
            else:
                if args.list_jobs:
                    restrict_headings = [
                        "PARENT_JOB_ID",
                        "VERB",
                        "EXECUTION_TIME",
                        "PROCESSING",
                        "SUCCESS",
                        "FAILED",
                    ]
                else:
                    restrict_headings = [
                        "EXPIRY",
                        "VERSION",
                        "HOST_UPTIME",
                        "AGENT_UPTIME",
                    ]
                (
                    tabulated_data,
                    headings,
                    computed_values,
                ) = _mixin.return_tabulated_data(
                    data=data, restrict_headings=restrict_headings
                )

        if tabulated_data:
            utils.print_tabulated_data(
                data=[i for i in tabulated_data if i], headers=headings
            )
            print("\nTotal Items: {}".format(len(tabulated_data)))
            for k, v in computed_values.items():
                if isinstance(v, float):
                    print("Total {}: {:.2f}".format(k, v))
                else:
                    print("Total {}: {}".format(k, v))
            else:
                return
    elif args.mode == "bootstrap":
        _bootstrap = bootstrap.Bootstrap(args=args)
        _bootstrap.bootstrap_cluster()
    else:
        parser.print_help(sys.stderr)
        raise SystemExit("Mode is set to an unsupported value.")



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import traceback

import yaml

from directord import components


class Component(components.ComponentBase):
    def __init__(self):
        """Initialize the component cachefile class.

        This component is not cacheable.
        """

        super().__init__(desc="Process cachefile commands")
        self.cacheable = False
        self.requires_lock = True

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "cachefile",
            help="Load a cached file and store it as an update to ARGs.",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        data["cachefile"] = self.known_args.cachefile
        return data

    def client(self, cache, job):
        """Run cache file operation.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        self.log.debug("client(): job: %s, cache: %s", job, cache)
        try:
            with open(job["cachefile"]) as f:
                cachefile_args = yaml.safe_load(f)
        except Exception as e:
            self.log.critical(str(e))
            return None, traceback.format_exc(), False, None
        else:
            self.set_cache(
                cache=cache,
                key="args",
                value=cachefile_args,
                value_update=True,
                tag="args",
                extend=True,
            )
            return "Cache file loaded", None, True, None



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import time

from directord import components


class Component(components.ComponentBase):
    def __init__(self):
        super().__init__(desc="Process query_wait commands")
        self.cacheable = False

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "item",
            help=(
                "Wait for an item to be part of the query cache. An item is"
                " a keys within the host specific query cache."
            ),
        )
        self.parser.add_argument(
            "--query-timeout",
            help=(
                "Wait for %(default)s seconds for a given item to be present"
                " in cache."
            ),
            default=600,
            type=int,
        )
        self.parser.add_argument(
            "--identity",
            help=(
                "Worker identities to search for a specific query item."
                " When an identity is defined, the lookup will search the"
                " defined identities for the item. QUERY_WAIT will block"
                " until the `item` is found in all specified identities."
                " this value can be used multiple times to express many"
                " identities."
            ),
            metavar="STRING",
            nargs="+",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted cacheevict action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        data["item"] = self.known_args.item
        data["query_timeout"] = self.known_args.query_timeout
        data["identity"] = self.known_args.identity
        return data

    def client(self, cache, job):
        """Run cache query_wait command operation.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        warning_loops = start_time = time.time()
        missing_identity = set()
        while (time.time() - start_time) < job["query_timeout"]:
            args = cache.get("args")
            if not args:
                continue

            query_args = args.get("query", dict())
            if not query_args:
                continue
            elif "identity" in job and job["identity"]:
                for identity in job["identity"]:
                    items = query_args.get(identity)
                    if isinstance(items, dict):
                        if job["item"] not in items.keys():
                            missing_identity.add(identity)
                            break
                        else:
                            if identity in missing_identity:
                                missing_identity.remove(identity)
                    else:
                        missing_identity.add(identity)
                        break
                else:
                    return (
                        "Item found in all identities",
                        None,
                        True,
                        (
                            "Item {} found in the query cache for"
                            " identities {}".format(
                                job["item"], job["identity"]
                            )
                        ),
                    )
                if missing_identity:
                    self.log.warning(
                        "QUERY argument [ %s ] not found in cache for %s",
                        job["item"],
                        missing_identity,
                    )
            else:
                for value in query_args.values():
                    if job["item"] in value:
                        return (
                            "Item found",
                            None,
                            True,
                            "Item {} found in the query cache".format(
                                job["item"]
                            ),
                        )

            if time.time() - warning_loops >= 5:
                self.log.debug(
                    "QUERY argument [ %s ] not found in cache", job["item"]
                )
                warning_loops = time.time()

            self.delay(0.01)

        if missing_identity:
            info = (
                "Item {} was not found in the query cache for the identities"
                " {} within {} seconds".format(
                    job["item"], list(missing_identity), job["query_timeout"]
                )
            )
        else:
            info = (
                "Item {} was not found in the query cache within"
                " {} seconds".format(job["item"], job["query_timeout"])
            )

        return (
            None,
            "Timeout after {} seconds".format(job["query_timeout"]),
            False,
            info,
        )



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from directord import components


class Component(components.ComponentBase):
    command = "arg"

    def __init__(self):
        """Initialize the component cache class.

        This component is not cacheable.
        """

        super().__init__(desc="Process cache commands")
        self.cacheable = False
        self.requires_lock = True

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "queue_sentinel",
            help="Activate the queue sentinel",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        data["parent_async_bypass"] = True
        return data

    def client(self, cache, job):
        """Run cache command operation.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :param command: Work directory path.
        :type command: String
        :returns: tuple
        """

        self.queue_sentinel = True
        return None, None, True, None



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import collections
import datetime
import json
import os
import time

import diskcache

import directord
from directord import components
from directord import interface
from directord import utils


class Client(interface.Interface):
    """Directord client class."""

    def __init__(self, args):
        """Initialize the Directord client.

        Sets up the client object.

        :param args: Arguments parsed by argparse.
        :type args: Object
        """

        super(Client, self).__init__(args=args)

        self.q_return = self.get_queue()
        self.base_component = components.ComponentBase()
        self.cache = dict()
        self.start_time = time.time()

    def q_processor(self, queue, lock):
        """Process jobs from the known queue.

        :param queue: Multiprocessing queue object.
        :type queue: Object
        :param lock: Locking object, used if a component requires it.
        :type lock: Object
        """

        while True:
            try:
                (component_kwargs, command, info) = queue.get(timeout=0.5)
            except ValueError as e:
                self.log.critical("Queue object value error [ %s ]", str(e))
                break
            except Exception:
                break
            else:
                self.log.debug(
                    "Job received [ %s ]", component_kwargs["job"]["job_id"]
                )
                self.job_q_component_run(component_kwargs, command, info, lock)

    def job_q_processor(self, q_processes, lock=None):
        """Process a given work queue.

        The defined `queue` is processed. The `processes` arg allows this
        method to spawn N processes.

        :param q_processes: Queue object
        :type q_processes: Object
        :param lock: Locking object, used if a component requires it.
        :type lock: Object
        """

        def _get_pending_bypass_threads(parents):
            """Return the current non-bypass thread objects.

            :param parents: Dictionary of all parent objects.
            :type parents: Dictionary
            :returns: List
            """

            return [
                i["t"].name
                for i in parents.values()
                if i["bypass"]
                and not i["t"].is_alive()
                and i["t"].exitcode is None
            ]

        def _get_pending_threads(parents):
            """Return the current non-bypass thread objects.

            :param parents: Dictionary of all parent objects.
            :type parents: Dictionary
            :returns: List
            """

            return [
                i["t"].name
                for i in parents.values()
                if not i["bypass"]
                and not i["t"].is_alive()
                and i["t"].exitcode is None
            ]

        def _get_thread_count(parents):
            """Return the current non-bypass thread count.

            :param parents: Dictionary of all parent objects.
            :type parents: Dictionary
            :returns: Integer
            """

            return len(
                [
                    True
                    for i in parents.values()
                    if not i["bypass"] and i["t"].is_alive()
                ]
            )

        if not lock:
            lock = self.get_lock()

        loop_time = time.time()
        parent_tracker = collections.OrderedDict()
        while not q_processes.empty() or parent_tracker:
            try:
                (
                    component_kwargs,
                    command,
                    info,
                ) = q_processes.get_nowait()
            except ValueError as e:
                self.log.critical("Queue object value error [ %s ]", str(e))
                pass
            except Exception:
                sleep_interval = 0.1
            else:
                sleep_interval = 0.001
                lower_command = command.decode().lower()
                if not hasattr(self, "__lock_{}__".format(lower_command)):
                    self.log.debug("Creating a new lock for %s", lower_command)
                    setattr(
                        self,
                        "__lock_{}__".format(lower_command),
                        self.get_lock(),
                    )

                job = component_kwargs["job"]
                self.log.debug("Received job_id [ %s ]", job["job_id"])
                # NOTE(cloudnull): If the command is queuesentinel purge all
                #                  queued items. This is on the ONE component
                #                  where we intercept and react outside of
                #                  the component structure.
                if lower_command == "queuesentinel":
                    count = 0
                    for value in list(parent_tracker.values()):
                        count += self.purge_queue(
                            queue=value["q"], job_id=job["job_id"]
                        )
                    self.log.info(
                        "Purged %s items from the work queues", count
                    )

                if job.get("parent_async_bypass") is True:
                    _q_name = "q_bypass_{}".format(
                        job.get("parent_id", job["job_id"])
                    )
                elif job.get("parent_async") is True:
                    _q_name = "q_async_{}".format(
                        job.get("parent_id", job["job_id"])
                    )
                else:
                    _q_name = "q_general"

                if _q_name in parent_tracker:
                    _parent = parent_tracker[_q_name]
                else:
                    _parent = parent_tracker[_q_name] = dict(
                        t=None,
                        q=self.get_queue(),
                        bypass=job.get("parent_async_bypass", False),
                    )
                    self.log.info("Parent queue [ %s ] created.", _q_name)

                _parent["q"].put((component_kwargs, command, info))
                if not _parent["t"]:
                    _parent["t"] = self.thread(
                        target=self.q_processor,
                        kwargs=dict(
                            queue=_parent["q"],
                            lock=lock,
                        ),
                        name=_q_name,
                        daemon=True,
                    )

            for t in _get_pending_bypass_threads(parents=parent_tracker):
                self.log.info("Starting bypass process [ %s ]", t)
                parent_tracker[t]["t"].start()

            while _get_thread_count(parents=parent_tracker) <= self.cpu_count:
                for t in _get_pending_threads(parents=parent_tracker):
                    self.log.info("Starting process [ %s ]", t)
                    parent_tracker[t]["t"].start()
                else:
                    break

            for key, value in list(parent_tracker.items()):
                if value["t"].is_alive() or value["t"].exitcode is None:
                    continue
                else:
                    if value["q"].empty():
                        timestamp = time.time()
                        if timestamp - value.get("timeout", timestamp) > 3:
                            self.terminate_process(process=value["t"])
                            value["q"].close()
                            value["q"].join_thread()
                            parent_tracker.pop(key)
                            self.log.info("Pruned parent [ %s ]", key)
                        else:
                            if "timeout" not in parent_tracker[key]:
                                self.log.info(
                                    "Starting to prune parent [ %s ]", key
                                )
                                parent_tracker[key]["timeout"] = timestamp
                    else:
                        self.log.info(
                            "Parent [ %s ] had items left in the queue,"
                            " respawning",
                            key,
                        )
                        parent_tracker[key].pop("timeout", None)
                        parent_tracker[key]["t"] = self.thread(
                            target=self.q_processor,
                            kwargs=dict(
                                queue=parent_tracker[key]["q"],
                                lock=lock,
                            ),
                            name=key,
                            daemon=True,
                        )

            if time.time() > loop_time:
                known_parents = list(parent_tracker.keys())
                if known_parents:
                    self.log.info("Known parents: %s", known_parents)

                self.log.info(
                    "Active processes: %s",
                    _get_thread_count(parents=parent_tracker),
                )
                loop_time = time.time() + 30

            time.sleep(sleep_interval)

        for component_lock in [
            i for i in self.__dict__.keys() if i.startswith("__lock_")
        ]:
            try:
                delattr(self, component_lock)
            except AttributeError:
                pass
            else:
                self.log.debug(
                    "Cleaned up removed dynamic component lock %s",
                    component_lock,
                )

    def purge_queue(self, queue, job_id):
        """Purge all jobs from the queue.

        :param queue: Queue object
        :type queue: Object
        :param job_id: Job UUID
        :type job_id: String
        :returns: Integer
        """

        total_count = 0
        while not queue.empty():
            try:
                (
                    _kwargs,
                    _command,
                    _,
                    _,
                ) = queue.get_nowait()
                total_count += 1
            except ValueError as e:
                self.log.critical("Queue object value error [ %s ]", str(e))
                break
            except Exception:
                break
            else:
                self.q_return.put(
                    (
                        None,
                        None,
                        False,
                        "Omitted due to sentinel from {}".format(job_id),
                        _kwargs["job"],
                        _command,
                        0,
                        None,
                    )
                )

        self.log.info("Cleared %s items from the work queue", total_count)
        return total_count

    def job_q_component_run(self, component_kwargs, command, info, lock):
        """Execute a component operation.

        Components are dynamically loaded based on the given component name.
        Upon execution, the results are put into the results queue.

        :param component_kwargs: Named arguments used with the componenet
                                 client.
        :type component_kwargs: Dictionary
        :param command: Byte encoded command used to run a given job.
        :type command: Bytes
        :param info: Information that was sent over with the original message.
        :type info: Bytes
        :param lock: Locking object, used if a component requires it.
        :type lock: Object
        """

        job = component_kwargs["job"]
        job_id = job["job_id"]
        command_lower = command.decode().lower()
        success, _, component = directord.component_import(
            component=command_lower,
            job_id=job_id,
        )

        cached = self.cache.get(
            job["job_sha3_224"]
        ) == self.driver.job_end.decode() and not job.get(
            "skip_cache", job.get("ignore_cache", False)
        )

        if not success:
            self.log.warning("Component lookup failure [ %s ]", component)
            self.q_return.put(
                (
                    None,
                    None,
                    success,
                    None,
                    job,
                    command,
                    0,
                    None,
                )
            )
        elif cached and component.cacheable is True:
            self.log.info(
                "Cache hit on [ %s ], task skipped.",
                job_id,
            )
            self.q_return.put(
                (
                    None,
                    None,
                    "skipped",
                    None,
                    job,
                    command,
                    0,
                    None,
                )
            )
        else:
            self.log.debug(
                "Starting component execution for job [ %s ].",
                job_id,
            )
            # Set the comment command argument
            setattr(component, "command", command)
            setattr(component, "info", info)
            setattr(component, "driver", self.driver)

            locked = False
            if component.requires_lock:
                try:
                    lock = getattr(self, "__lock_{}__".format(command_lower))
                except AttributeError:
                    self.log.warning(
                        "No component lock found for [ %s ], falling back"
                        " to global lock",
                        command_lower,
                    )
                else:
                    self.log.debug(
                        "Found component lock [ %s ]",
                        "__lock_{}__".format(command_lower),
                    )

                locked = lock.acquire()
                self.log.debug("Lock acquired for [ %s ]", job_id)

            with self.timeout(
                time=job.get("timeout", 600),
                job_id=job_id,
            ):
                _starttime = time.time()
                stdout, stderr, outcome, info = component.client(
                    cache=self.cache, job=job
                )
                job[
                    "component_exec_timestamp"
                ] = datetime.datetime.fromtimestamp(time.time()).strftime(
                    "%Y-%m-%d %H:%M:%S"
                )

            try:
                if component.block_on_tasks:
                    block_on_tasks_data = [
                        i
                        for i in component.block_on_tasks
                        if self.driver.identity in i.get("targets", list())
                    ]
                    if outcome and block_on_tasks_data:
                        outcome = None
                        info = "Waiting for callback tasks to complete"
                else:
                    block_on_tasks_data = None

                component_return = (
                    stdout,
                    stderr,
                    outcome,
                    info,
                    job,
                    command,
                    time.time() - _starttime,
                    component.block_on_tasks,
                )
            except UnboundLocalError:
                component.block_on_tasks = list()
                component_return = (
                    None,
                    None,
                    False,
                    "Job was unable to finish",
                    job,
                    command,
                    time.time() - _starttime,
                    None,
                )

            if locked:
                lock.release()
                self.log.debug("Lock released for [ %s ]", job_id)

            self.q_return.put(component_return)

            try:
                block_on_task_data = block_on_tasks_data[-1]
            except IndexError:
                self.log.debug(
                    "Job [ %s ] no valid callbacks for this node %s.",
                    job["job_id"],
                    self.driver.identity,
                )
            except TypeError:
                self.log.debug("No callbacks defined.")
            else:
                self.log.info(
                    "Job [ %s ] number of job call backs [ %s ]",
                    job["job_id"],
                    len(component.block_on_tasks),
                )
                self.log.debug("Job call backs: %s ", component.block_on_tasks)
                block_on_task_success = False
                with self.timeout(
                    time=block_on_task_data.get("timeout", 600),
                    job_id=block_on_task_data["job_id"],
                ):
                    while True:
                        if self.cache.get(
                            block_on_task_data["job_sha3_224"]
                        ) in [
                            self.driver.job_end.decode(),
                            self.driver.job_failed.decode(),
                        ]:
                            block_on_task_success = True
                            break
                        else:
                            self.log.debug(
                                "waiting for callback job from [ %s ] to"
                                " complete. %s",
                                job["job_id"],
                                block_on_task_data,
                            )
                            time.sleep(1)

                if block_on_task_success:
                    self.log.debug(
                        "Job [ %s ] task sha [ %s ] callback complete",
                        job["job_id"],
                        block_on_task_data["job_sha3_224"],
                    )
                    self.q_return.put(
                        (
                            stdout,
                            stderr,
                            True,
                            "Callback [ {} ] completed".format(
                                block_on_task_data["job_id"]
                            ),
                            job,
                            command,
                            time.time() - _starttime,
                            None,
                        )
                    )
                else:
                    self.log.error(
                        "Job [ %s ] task sha [ %s ] callback never"
                        " completed",
                        job["job_id"],
                        block_on_task_data["job_sha3_224"],
                    )
                    self.q_return.put(
                        (
                            stdout,
                            stderr,
                            False,
                            "Callback [ {} ] never completed".format(
                                block_on_task_data["job_id"]
                            ),
                            job,
                            command,
                            time.time() - _starttime,
                            None,
                        )
                    )

            self.log.debug(
                "Component execution complete for job [ %s ].",
                job["job_id"],
            )

    def _set_job_status(
        self, stdout, stderr, outcome, return_info, job, block_on_tasks, conn
    ):
        """Set job status.

        :param stdout: Execution standard output.
        :type stdout: String|Bytes
        :param stderr: Execution standard error.
        :type stderr: String|Bytes
        :param outcome: Outcome status information.
        :type outcome: Boolean|String
        :param return_info: Information returning from component execution.
        :type return_info: Bytes
        :param job: Job definition
        :type job: Dictionary
        :param block_on_tasks: Job to post back to the server to create
                               a new task which the client will wait
                               for.
        :type block_on_tasks: List
        :param conn: Job bind connection object.
        :type conn: Object
        """

        if stdout:
            stdout = stdout.strip()
            if not isinstance(stdout, bytes):
                stdout = stdout.encode()
            conn.stdout = stdout
            self.log.debug("Job [ %s ], stdout: %s", job["job_id"], stdout)

        if stderr:
            stderr = stderr.strip()
            if not isinstance(stderr, bytes):
                stderr = stderr.encode()
            conn.stderr = stderr
            self.log.warning("Job [ %s ], stderr: %s", job["job_id"], stderr)

        if outcome is False:
            state = conn.job_state = self.driver.job_failed
            self.log.error("Job failed [ %s ]", job["job_id"])
        elif outcome is True:
            state = conn.job_state = self.driver.job_end
            self.log.info("Job complete [ %s ]", job["job_id"])
        elif outcome == "skipped":
            self.log.info("Job skipped [ %s ]", job["job_id"])
            conn.info = b"task skipped"
            state = conn.job_state = self.driver.job_end
        else:
            state = conn.job_state = self.driver.job_processing

        conn.info = b"task finished"
        if return_info:
            conn.info = return_info

        if block_on_tasks:
            job["new_tasks"] = block_on_tasks

        job["return_timestamp"] = datetime.datetime.fromtimestamp(
            time.time()
        ).strftime("%Y-%m-%d %H:%M:%S")

        if "new_tasks" in job:
            conn.data = json.dumps(job).encode()
        else:
            minimal_data = {
                "execution_time": job["execution_time"],
                "return_timestamp": job["return_timestamp"],
            }
            component_timestamp = job.get("component_exec_timestamp")
            if component_timestamp:
                minimal_data["component_exec_timestamp"] = component_timestamp

            conn.data = json.dumps(minimal_data).encode()

        if job["parent_id"]:
            self.base_component.set_cache(
                cache=self.cache,
                key=job["parent_id"],
                value=state,
                tag="parents",
            )

        self.base_component.set_cache(
            cache=self.cache,
            key=job["job_sha3_224"],
            value=state,
            tag="jobs",
        )

    def job_q_results(self):
        """Job results queue processor.

        Results are retrieved from the queue and status is set.
        """

        try:
            (
                stdout,
                stderr,
                outcome,
                return_info,
                job,
                command,
                execution_time,
                block_on_tasks,
            ) = self.q_return.get_nowait()
        except ValueError as e:
            self.log.critical("Return object value error [ %s ]", str(e))
        except Exception:
            pass
        else:
            self.log.debug("Found task results for [ %s ].", job["job_id"])
            with utils.ClientStatus(
                socket=self.bind_job,
                job_id=job["job_id"].encode(),
                command=command,
                ctx=self,
            ) as c:
                job["execution_time"] = execution_time
                self._set_job_status(
                    stdout,
                    stderr,
                    outcome,
                    return_info,
                    job,
                    block_on_tasks,
                    c,
                )

    def _parent_check(self, conn, cache, job):
        """Check if a parent job has failed.

        This will check if tasks under a given parent job are allowed to
        continue executing.

        :param conn: Job bind connection object.
        :type conn: Object
        :param cache: Caching object used to save information to the local
                      client.
        :type cache: Object
        :param job: Job definition
        :type job: Dictionary

        :returns: Boolean
        """
        if cache.get(job["parent_id"]) == self.driver.job_failed.decode():
            self.log.error(
                "Parent failure %s skipping %s",
                job["parent_id"],
                job["job_id"],
            )
            status = (
                "Job [ {} ] was not allowed to run because"
                " there was a failure under this partent ID"
                " [ {} ]".format(job["job_id"], job["parent_id"])
            )

            self.log.error(status)
            conn.info = status.encode()
            conn.job_state = self.driver.job_failed
            return False
        else:
            return True

    def prune_cache(self, cache_check_time):
        """Prune the local cache to ensure a tidy environment.

        If there are any warnings when pruning, they will be logged.

        :param cache_check_time: Timestamp from last prune.
        :type cache_check_time: Float
        :returns: Float
        """

        if time.time() > cache_check_time + 4096:
            self.log.info(
                "Current estimated cache size: %s KiB",
                self.cache.volume() / 1024,
            )

            warnings = self.cache.check()
            if warnings:
                self.log.warning(
                    "Client cache noticed %s warnings.", len(warnings)
                )
                for item in warnings:
                    self.log.warning(
                        "Client Cache Warning: [ %s ].",
                        str(item.message),
                    )

            self.cache.expire()
            return time.time()

        return cache_check_time

    def run_job(self, lock=None, sentinel=False):
        """Job entry point.

        This creates a cached access object, connects to the socket and begins
        the loop.

        > When a file transfer is initiated the client will enter a loop
          waiting for data chunks until an `transfer_end` signal is passed.

        * Initial poll interval is 1024, maxing out at 2048. When work is
          present, the poll interval is 1.

        :param sentinel: Breaks the loop
        :type sentinel: Boolean
        """

        self.bind_job = self.driver.job_connect()
        poller_time = time.time()
        heartbeat_time = time.time()
        poller_interval = 1
        cache_check_time = time.time()
        run_q_processor_thread = None
        q_processes = self.get_queue()
        while True:
            if self.terminate_process(process=run_q_processor_thread):
                run_q_processor_thread = None

            if not q_processes.empty() and not run_q_processor_thread:
                run_q_processor_thread = self.thread(
                    target=self.job_q_processor,
                    kwargs=dict(q_processes=q_processes, lock=lock),
                    name="job_q_processor",
                    daemon=False,
                )
                run_q_processor_thread.start()

            self.job_q_results()
            if self.q_return.empty():
                poller_interval = utils.return_poller_interval(
                    poller_time=poller_time,
                    poller_interval=poller_interval,
                    log=self.log,
                )

            if time.time() > heartbeat_time:
                with open("/proc/uptime", "r") as f:
                    uptime = float(f.readline().split()[0])

                self.driver.socket_send(
                    socket=self.bind_job,
                    control=self.driver.heartbeat_notice,
                    data=json.dumps(
                        {
                            "version": directord.__version__,
                            "host_uptime": str(
                                datetime.timedelta(seconds=uptime)
                            ),
                            "agent_uptime": str(
                                datetime.timedelta(
                                    seconds=(time.time() - self.start_time)
                                )
                            ),
                        }
                    ).encode(),
                )
                heartbeat_time = time.time() + 30
                self.log.info("Heartbeat sent to server")

            if self.driver.bind_check(
                bind=self.bind_job, constant=poller_interval
            ):
                poller_interval, poller_time = 1, time.time()
                (
                    _,
                    _,
                    command,
                    data,
                    info,
                    _,
                    _,
                ) = self.driver.socket_recv(socket=self.bind_job)
                job = json.loads(data.decode())
                job["job_id"] = job_id = job.get("job_id", utils.get_uuid())
                job["job_sha3_224"] = job_sha3_224 = job.get(
                    "job_sha3_224", utils.object_sha3_224(job)
                )
                self.driver.socket_send(
                    socket=self.bind_job,
                    msg_id=job_id.encode(),
                    control=self.driver.job_ack,
                )

                job_parent_id = job.get("parent_id")
                job_parent_sha3_224 = job.get("parent_sha3_224")
                self.log.info(
                    "Item received: parent job UUID [ %s ],"
                    " parent job sha3_224 [ %s ], job UUID [ %s ],"
                    " job sha3_224 [ %s ]",
                    job_parent_id,
                    job_parent_sha3_224,
                    job_id,
                    job_sha3_224,
                )

                with utils.ClientStatus(
                    socket=self.bind_job,
                    job_id=job_id.encode(),
                    command=command,
                    ctx=self,
                ) as c:
                    if job_parent_id and not self._parent_check(
                        conn=c, cache=self.cache, job=job
                    ):
                        self.q_return.put(
                            (
                                None,
                                None,
                                False,
                                b"Job omitted, parent failure",
                                job,
                                command,
                                0,
                                None,
                            )
                        )
                    else:
                        c.job_state = self.driver.job_processing
                        component_kwargs = dict(cache=None, job=job)
                        self.log.debug(
                            "Queuing component [ %s ], job_id [ %s ]",
                            command.decode(),
                            job_id,
                        )
                        c.info = b"task queued"
                        q_processes.put(
                            (
                                component_kwargs,
                                command,
                                info,
                            )
                        )
            else:
                cache_check_time = self.prune_cache(
                    cache_check_time=cache_check_time
                )
                time.sleep(poller_interval * 0.001)

            if sentinel:
                break

    def worker_run(self):
        """Run all work related threads.

        Threads are gathered into a list of process objects then fed into the
        run_threads method where their execution will be managed.
        """

        lock = self.get_lock()
        threads = [
            (
                self.thread(
                    name="run_job", target=self.run_job, kwargs=dict(lock=lock)
                ),
                False,
            ),
        ]
        # Ensure that the cache path exists before executing.
        os.makedirs(self.args.cache_path, exist_ok=True)
        with diskcache.Cache(
            self.args.cache_path,
            tag_index=True,
            disk=diskcache.JSONDisk,
        ) as cache:
            self.cache = cache
            self.run_threads(threads=threads)



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging
import os

from logging import handlers


def getLogger(name, debug_logging=False):
    """Return a logger from a given name.

    If the name does not have a log handler, this will create one for it based
    on the module name which will log everything to a log file in a location
    the executing user will have access to.

    :param name: Log handler name to retrieve.
    :type name: String
    :returns: Object
    """

    log = logging.getLogger(name=name)
    for handler in log.handlers:
        if name == handler.name:
            if debug_logging:
                handler.setLevel(logging.DEBUG)
                log.setLevel(logging.DEBUG)
            return log
    else:
        return LogSetup(debug_logging=debug_logging).default_logger(
            name=name.split(".")[0]
        )


class LogSetup:
    """Logging Class."""

    def __init__(self, max_size=500, max_backup=5, debug_logging=False):
        """Setup Logging.

        :param max_size: Set max log file size
        :type max_size: Integer
        :param max_backup: Set max log file back rotations.
        :type max_backup: Integer
        :param debug_logging: Enable | Disable debug logging.
        :type debug_logging: Boolean
        """

        self.max_size = max_size * 1024 * 1024
        self.max_backup = max_backup
        self.debug_logging = debug_logging
        self.format = None
        self.name = None
        self.enable_stream = False
        self.enable_file = False

    def default_logger(
        self, name=__name__, enable_stream=True, enable_file=False
    ):
        """Default Logger.

        This is set to use a rotating File handler and a stream handler.
        If you use this logger all logged output that is INFO and above will
        be logged, unless debug_logging is set then everything is logged.
        The logger will send the same data to a stdout as it does to the
        specified log file.

        You can disable the default handlers by setting either `enable_file` or
        `enable_stream` to `False`

        :param name: Log handler name to retrieve.
        :type name: String
        :param enable_stream: Enable | Disable log Streaming.
        :type enable_stream: Boolean
        :param enable_file: Enable | Disable log writting to a file.
        :type enable_file: Boolean
        :returns: Object
        """

        log = logging.getLogger(name)
        self.name = name

        if enable_file:
            self.enable_file = enable_file
            file_handler = handlers.RotatingFileHandler(
                filename=self.return_logfile(filename="%s.log" % name),
                maxBytes=self.max_size,
                backupCount=self.max_backup,
            )
            self.set_handler(log, handler=file_handler)

        if enable_stream:
            self.enable_stream = enable_stream
            stream_handler = logging.StreamHandler()
            self.set_handler(log, handler=stream_handler)

        return log

    def set_handler(self, log, handler):
        """Set the logging level as well as the handlers.

        :param log: Logging object.
        :type log: Object
        :param handler: Log handler object.
        :type handler: Object
        """

        if self.debug_logging:
            log.setLevel(logging.DEBUG)
            handler.setLevel(logging.DEBUG)
        else:
            log.setLevel(logging.INFO)
            handler.setLevel(logging.INFO)

        handler.name = self.name

        if not self.format:
            if self.enable_file:
                self.format = logging.Formatter(
                    "%(asctime)s %(levelname)s %(message)s"
                )
            else:
                self.format = logging.Formatter("%(levelname)s %(message)s")

        handler.setFormatter(self.format)
        log.addHandler(handler)

    @staticmethod
    def return_logfile(filename, log_dir="/var/log"):
        """Return a path for logging file.

        If `log_dir` exists and the userID is 0 the log file will be written
        to the provided log directory. If the UserID is not 0 or log_dir does
        not exist the log file will be written to the users home folder.

        :param filename: File name to write log messages.
        :type filename: String
        :param log_dir: Directory where the log file will be stored.
        :type log_dir: String
        :returns: String
        """

        user = os.getuid()
        home = os.path.expanduser("~")

        if not os.path.isdir(log_dir):
            return os.path.join(home, filename)

        log_dir_stat = os.stat(log_dir)
        if log_dir_stat.st_uid == user:
            return os.path.join(log_dir, filename)
        elif log_dir_stat.st_gid == user:
            return os.path.join(log_dir, filename)

        return os.path.join(home, filename)



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import contextlib
import importlib
import importlib.util as importlib_util
import json
import multiprocessing
import os
import signal
import socket
import sys

from types import SimpleNamespace

from directord import meta
from directord import logger
from directord.meta import __version__  # noqa


def send_data(socket_path, data):
    """Send data to the socket path.

    The send method takes serialized data and submits it to the given
    socket path.

    This method will return information provided by the server in
    String format.

    :returns: String
    """

    try:
        with UNIXSocketConnect(socket_path) as s:
            if not s:
                raise SystemExit("No connection available to server.")
            s.sendall(data.encode())
            fragments = []
            while True:
                chunk = s.recv(1024)
                if not chunk:
                    break

                fragments.append(chunk)
            return b"".join(fragments)
    except PermissionError:
        log = logger.getLogger(name="directord")
        error_msg = (
            "Permission error writing to {}. Check write permissions.".format(
                socket_path
            )
        )
        log.error(
            error_msg,
        )
        raise PermissionError(error_msg)


def plugin_import(plugin):
    """Import a plugin from string.

    :param plugin: Python import in dot notation.
    :type plugin: String
    :returns: Object
    """

    return importlib.import_module(plugin, package="directord")


def component_import(component, job_id=None):
    """Import a component and return a tuple with the class object.

    If the component isn't a builtin the system will search
    the shared path for a user defined component.

    > Return: (Boolean, Boolean, Object|String)

    :param component: String name of the component.
    :type component: String
    :param job_id: Job UUID, used client side.
    :type job_id: String
    :returns: Tuple
    """

    try:
        component_obj = plugin_import(
            plugin=".components.builtin_{}".format(component)
        )
        transfer = None
    except ImportError as e:
        error = str(e)
        paths = [
            os.path.join(sys.base_prefix, "share/directord/components"),
            "/etc/directord/components",
        ]
        if sys.base_prefix != sys.prefix:
            paths.insert(
                0, os.path.join(sys.prefix, "share/directord/components")
            )

        for path in paths:
            try:
                transfer = os.path.join(path, "{}.py".format(component))
                name = "directord_user_component_{}".format(component)
                spec = importlib_util.spec_from_file_location(name, transfer)
                component_obj = importlib_util.module_from_spec(spec)
                sys.modules[name] = component_obj
                spec.loader.exec_module(component_obj)
            except (ImportError, FileNotFoundError) as e:
                error += "\n{}".format(str(e))
            else:
                break
        else:
            info = (
                "Failure - Unknown Component\n"
                "ERROR:{}\nCOMMAND:{}\nID:{}\nPATH:{}".format(
                    error, component, job_id, paths
                )
            )
            return False, transfer, info

    return True, transfer, component_obj.Component()


class Processor:
    """Processing class, provides queing and threading utilities.

    This is a base class.
    """

    cpu_count = multiprocessing.cpu_count() * 2
    if cpu_count > 16:
        cpu_count = 16

    thread = multiprocessing.Process

    def __init__(self):
        """Initialize Processor class creating all required manager objects.

        Managers maintain a multiprocessing proxy object which allows data to
        be shared across threads.
        """

        self.log = logger.getLogger(name="directord")

    @staticmethod
    def get_manager():
        """Returns a multiprocessing manager."""

        return multiprocessing.Manager()

    @staticmethod
    def get_lock():
        """Returns a multiprocessing lock."""

        return multiprocessing.Lock()

    @staticmethod
    def get_queue():
        """Returns a multiprocessing queue."""

        return multiprocessing.Queue()

    def terminate_process(self, process):
        """Terminate a processes when possible.

        * If the process object is null, return False
        * If the process is in an "alive", return False.

        :param process: Multiprocessing object.
        :type process: Object
        :returns: Boolean
        """

        if not process:
            return False

        try:
            if not process.is_alive():
                process.join(timeout=0.1)
                process.terminate()
                self.log.info("Process [ %s ] cleaned up", process.name)
                return True
        except Exception as e:
            self.log.warning(
                "Thread [ %s ] saw exception when terminating: %s",
                process.name,
                str(e),
            )

        return False

    def run_threads(self, threads):
        """Execute process objects from an array.

        The array of threads are processed and started in a "daemon" mode.
        Once started the thread object is added into a cleanup array which
        is then joined.

        > Each item within the threads list consists of a tuple
          (Process(), Boolean). The boolean is used to enable or disable
          a daemonic process.

        :param threads: An array of Process objects.
        :type threads: List
        """

        for t, daemon in threads:
            t.daemon = daemon
            t.start()

        for t, _ in threads:
            t.join()

    @contextlib.contextmanager
    def timeout(self, time, job_id, reraise=False):
        """Registers a context manager to raise whenever a timeout occurs.

        :param time: Time in seconds before an alarm is raised.
        :type time: Integer
        :param job_id: Job UUID
        :type job_id: String
        :param reraise: Reraise the timeout exception
        :type reraise: Boolean
        :yields:
        """

        signal.signal(signal.SIGALRM, self.raise_timeout)
        signal.alarm(time)
        try:
            yield
        except TimeoutError:
            self.log.warning(
                "Timeout encountered after [ %s ] seconds running [ %s ].",
                time,
                job_id,
            )
            if reraise:
                raise TimeoutError
        finally:
            signal.signal(signal.SIGALRM, signal.SIG_IGN)

    def raise_timeout(self, *args, **kwargs):
        """Log, then raise a Timeout error."""

        self.log.error("Task timeout encountered.")
        raise TimeoutError


class UNIXSocketConnect:
    """Context manager for connecting to a UNIX socket."""

    def __init__(self, sock_path):
        """Initialize the UNIX socket connect context manager.

        :param socket_path: Path to the local file system socket.
        :type socket_path: String
        """

        self.socket_path = sock_path
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

    def __enter__(self):
        """Upon enter, connect to the socket and return the socket object.

        :returns: Object
        """

        try:
            self.sock.connect(self.socket_path)
        except ConnectionRefusedError as e:
            print(str(e))
            self.__exit__()
        else:
            return self.sock

    def __exit__(self, *args, **kwargs):
        """Upon exit, close the unix socket."""

        self.sock.close()


class DirectordConnect:
    """Library context manager providing easy access into Directord."""

    def __init__(
        self,
        debug=False,
        socket_path="/var/run/directord.sock",
        driver=meta.__driver_default__,
        force_async=False,
    ):
        """Initialize the connection.

        Basic usage.

        > with DirectordConnect() as d:
        ...   # Run orchestrations.
        ...   ids = d.orchestrate(
        ...       orchestrations=jobs
        ...   )

        :param debug: Enable|Disable debug mode.
        :type debug: Boolean
        :param socket_path: Socket path used to connect to Directord.
        :type socket_path: String
        :param force_async: Forces all orchestrations to run asynchronously.
        :type: force_async: Boolean
        """

        args = SimpleNamespace(
            **{
                "debug": debug,
                "socket_path": socket_path,
                "driver": driver,
                "force_async": force_async,
            }
        )
        _mixin = plugin_import(plugin=".mixin")
        self.mixin = _mixin.Mixin(args=args)
        _user = plugin_import(plugin=".user")
        self.manage = _user.Manage(args=args)

    def __enter__(self):
        """Enter the context manager returning self."""

        return self

    def __exit__(self, *args, **kwargs):
        """Shutdown the context manager."""

        pass

    @staticmethod
    def _from_json(return_obj):
        """Decode a byte object and return the loaded JSON data.

        :param return_obj: Byte encoded JSON string.
        :type return_objL Bytes
        :returns: Dictionary
        """

        return json.loads(return_obj.decode())

    def orchestrate(self, orchestrations, defined_targets=None):
        """Run an orchestration and return a list of job IDs.

        :param orchestrations: List of dictionary objects used to run
                               orchestrations.
        :type orchestrations: List
        :param defined_targets: List of Directord Targets.
        :type defined_targets: List
        :returns: List
        """

        return [
            i.decode()
            for i in self.mixin.exec_orchestrations(
                orchestrations,
                defined_targets=defined_targets,
                return_raw=True,
            )
        ]

    def poll(self, job_id):
        """Poll for the completion of a given job ID.

        :param job_id: Job UUID.
        :type job_id: String
        :returns: Tuple
        """

        state, status, _, _, _ = self.manage.poll_job(job_id=job_id)
        return state, status

    def list_nodes(self):
        """Return a list of all active Directord Nodes.

        :returns: List
        """

        return list(
            dict(
                self._from_json(self.manage.run(override="list-nodes"))
            ).keys()
        )

    def list_jobs(self):
        """Return a dictionary of all current Directord jobs.

        :returns: Dictionary
        """

        return dict(self._from_json(self.manage.run(override="list-jobs")))

    def purge_nodes(self):
        """Purge all nodes from the work pool.

        Purge all nodes from the pool, all remaining active nodes will
        recheck-in and be added to the pool.

        :returns: Boolean
        """

        return self._from_json(self.manage.run(override="purge-nodes"))[
            "success"
        ]

    def purge_jobs(self):
        """Purge all jobs from the return manager.

        Purge all jobs from the return manager.

        :returns: Boolean
        """

        return self._from_json(self.manage.run(override="purge-jobs"))[
            "success"
        ]



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import time
import queue

from directord import components


class Coordination:
    """Coordination connection context manager."""

    def __init__(self, driver, log, job_id):
        """Initialize the coordination context manager class."""

        self.driver = driver
        self.bind_backend = self.driver.backend_connect()
        self.log = log
        self.job_id = job_id

    def __enter__(self):
        """Return the bind_backend object on enter."""

        self.log.debug(
            "Coordination started to %s for job [ %s ]",
            self.driver.identity,
            self.job_id,
        )
        return self.bind_backend

    def __exit__(self, *args, **kwargs):
        """Close the bind coordination object."""

        try:
            self.bind_backend.close(linger=2)
            close_timeout = time.time()
            while not self.bind_backend.closed:
                if time.time() - close_timeout > 60:
                    raise TimeoutError(
                        "Job [ {} ] failed to close coordination"
                        " socket".format(self.job_id)
                    )
                else:
                    self.bind_backend.close(linger=2)
                    time.sleep(1)
        except Exception as e:
            self.log.error(
                "Job [ %s ] coordination ran into an exception"
                " while closing the socket %s",
                str(e),
                self.job_id,
            )
        else:
            self.log.debug(
                "Job [ %s ] coordination socket closed", self.job_id
            )

        self.log.debug(
            "Coordination ended for %s for job [ %s ]",
            self.driver.identity,
            self.job_id,
        )


class Component(components.ComponentBase):
    def __init__(self):
        """Initialize the component cache class."""

        super().__init__(desc="Process coordination commands")
        self.cacheable = False

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "sha",
            help=("job sha to be completed."),
        )
        self.parser.add_argument(
            "--identity",
            help=(
                "Worker identities to search for a specific job item."
                " When an identity is defined, the lookup will search the"
                " defined identities for the item. JOB_WAIT will block"
                " until the `item` is found in all specified identities."
                " this value can be used multiple times to express many"
                " identities."
            ),
            metavar="STRING",
            nargs="+",
            required=True,
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted cacheevict action.
        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        data["job_sha"] = self.known_args.sha
        data["identity"] = self.known_args.identity
        return data

    def client(self, cache, job):
        """Run file transfer operation.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        driver = self.driver.__copy__()
        self.log.debug("client(): job: %s, cache: %s", job, cache)
        with Coordination(
            driver=driver, log=self.log, job_id=job["job_id"]
        ) as bind_backend:
            return self._client(cache, job, driver, bind_backend)

    def _client(self, cache, job, driver, bind_backend):
        """Run cache query_wait command operation.
        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :param driver: Connection object used to store information used in a
                     return message.
        :type driver: Object
        :returns: tuple
        """

        if not job["identity"]:
            return (
                None,
                None,
                True,
                "No identities to process",
            )

        q = queue.Queue()

        if driver.identity in job["identity"]:
            job["identity"].remove(driver.identity)
            job["identity"].append(driver.identity)

        list(map(q.put, job["identity"]))
        confirmed_identities = set()
        all_identities_sent = False
        while True:
            try:
                identity = q.get_nowait()
            except queue.Empty:
                all_identities_sent = True
            else:
                driver.socket_send(
                    socket=bind_backend,
                    msg_id=job["job_id"].encode(),
                    control=driver.coordination_notice,
                    data=job["job_sha"].encode(),
                    info=identity.encode(),
                )
                self.log.debug(
                    "Job [ %s ] coordination notice sent to %s",
                    job["job_id"],
                    identity,
                )

            if driver.bind_check(bind=bind_backend, interval=0.5):
                (
                    msg_id,
                    control,
                    _,
                    data,
                    info,
                    stderr,
                    stdout,
                ) = driver.socket_recv(socket=bind_backend)

                if control == driver.coordination_notice:
                    if data != job["job_sha"].encode():
                        self.log.debug(
                            "Job [ %s ] coordination notice received from"
                            " [ %s ] but data miss-matched, sending back.",
                            msg_id.decode(),
                            info.decode(),
                        )
                        driver.socket_send(
                            msg_id=msg_id,
                            control=control,
                            data=data,
                            info=info,
                            stderr=stderr,
                            stdout=stdout,
                        )
                    else:
                        self.log.debug(
                            "Job [ %s ] coordination notice received from"
                            " [ %s ]",
                            msg_id.decode(),
                            info.decode(),
                        )
                        for _ in range(2400):
                            if cache.get(data.decode()) in [
                                driver.job_end.decode(),
                                driver.job_failed.decode(),
                            ]:
                                self.log.debug(
                                    "Job [ %s ] coordination complete for"
                                    " [ %s ]",
                                    msg_id.decode(),
                                    info.decode(),
                                )
                                driver.socket_send(
                                    socket=bind_backend,
                                    msg_id=msg_id,
                                    control=driver.coordination_ack,
                                    info=info,
                                )
                                break
                            self.delay(0.25)
                        else:
                            self.log.debug(
                                "Job [ %s ] expected SHA [ %s ] was not"
                                " found.",
                                msg_id.decode(),
                                data.decode(),
                            )
                            driver.socket_send(
                                socket=bind_backend,
                                msg_id=msg_id,
                                control=driver.coordination_failed,
                                info=info,
                                stderr=b"Item was not found in cache",
                            )
                elif control == driver.coordination_ack:
                    self.log.debug(
                        "Job [ %s ] coordination ACK for [ %s ] received",
                        msg_id.decode(),
                        info.decode(),
                    )
                    confirmed_identities.add(info.decode())
                elif control == driver.coordination_failed:
                    self.log.error(
                        "Job [ %s ] coordination failed from"
                        " [ %s ] error %s",
                        msg_id.decode(),
                        info.decode(),
                        data.decode(),
                    )
                    return (
                        stderr.decode(),
                        stdout.decode(),
                        False,
                        "Job [ {} ] failed when attempting coordination with"
                        " [ {} ]".format(msg_id.decode(), info.decode()),
                    )
                else:
                    self.log.critical(
                        "Unknown control received [ %s ] from [ %s ]",
                        control.decode(),
                        info.decode(),
                    )
            elif (
                sorted(confirmed_identities) == sorted(job["identity"])
                and all_identities_sent
            ):
                self.log.debug(
                    "Job [ %s ] coordination with %s success",
                    job["job_id"],
                    confirmed_identities,
                )
                return (
                    "Job completed, found SHA [ {} ]".format(job["job_sha"]),
                    None,
                    True,
                    "Job [ {} ] completed on all"
                    " coordinated targets: {}".format(
                        job["job_id"], job["identity"]
                    ),
                )
            else:
                self.log.debug(
                    "Waiting for coordination messages from %s",
                    sorted(set(job["identity"]) - confirmed_identities),
                )



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import collections
import json
import os
import time

import diskcache

import directord

from directord import interface


class User(interface.Interface):
    """Directord User interface class."""

    def __init__(self, args):
        """Initialize the User interface class.

        Sets up the user object.

        :param args: Arguments parsed by argparse.
        :type args: Object
        """

        super(User, self).__init__(args=args)


class Manage(User):
    """Directord Manage interface class."""

    def __init__(self, args):
        """Initialize the Manage interface class.

        Sets up the manage object.

        :param args: Arguments parsed by argparse.
        :type args: Object
        """

        super(User, self).__init__(args=args)

    @staticmethod
    def move_certificates(
        directory, target_directory=None, backup=False, suffix=".key"
    ):
        """Move certificates when required.

        :param directory: Set the origin path.
        :type directory: String
        :param target_directory: Set the target path.
        :type target_directory: String
        :param backup: Enable file backup before moving.
        :type backup:  Boolean
        :param suffix: Set the search suffix
        :type suffix: String
        """

        for item in os.listdir(directory):
            if backup:
                target_file = "{}.bak".format(os.path.basename(item))
            else:
                target_file = os.path.basename(item)

            if item.endswith(suffix):
                os.rename(
                    os.path.join(directory, item),
                    os.path.join(target_directory or directory, target_file),
                )

    def generate_certificates(self, base_dir="/etc/directord"):
        """Generate client and server CURVE certificate files.

        :param base_dir: Directord configuration path.
        :type base_dir: String
        """

        keys_dir = os.path.join(base_dir, "certificates")
        public_keys_dir = os.path.join(base_dir, "public_keys")
        secret_keys_dir = os.path.join(base_dir, "private_keys")

        for item in [keys_dir, public_keys_dir, secret_keys_dir]:
            os.makedirs(item, exist_ok=True)

        # Run certificate backup
        self.move_certificates(directory=public_keys_dir, backup=True)
        self.move_certificates(
            directory=secret_keys_dir, backup=True, suffix=".key_secret"
        )

        # create new keys in certificates dir
        for item in ["server", "client"]:
            self.driver.key_generate(keys_dir=keys_dir, key_type=item)

        # Move generated certificates in place
        self.move_certificates(
            directory=keys_dir,
            target_directory=public_keys_dir,
            suffix=".key",
        )
        self.move_certificates(
            directory=keys_dir,
            target_directory=secret_keys_dir,
            suffix=".key_secret",
        )

    def poll_job(self, job_id):
        """Given a job poll for its completion and return status.

        > The status return is (Boolean, String)

        :param job_id: UUID for job
        :type job_id: String
        :returns: Tuple
        """

        job_processing_interval = 0.25
        processing_attempts = 0
        state_timeout = time.time()
        timeout = getattr(self.args, "timeout", 600)
        while True:
            try:
                data = dict(json.loads(self.run(override=job_id)))
            except json.JSONDecodeError:
                if time.time() - state_timeout > timeout:
                    state_timeout = time.time()
                    return (
                        None,
                        "Job in an unknown state: {}".format(job_id),
                        None,
                        None,
                        None,
                    )
                else:
                    time.sleep(1)
                continue
            else:
                data_return = data.get(job_id, dict())
                if not data_return:
                    if time.time() - state_timeout > timeout:
                        state_timeout = time.time()
                        return (
                            None,
                            "Job in an unknown state: {}".format(job_id),
                            None,
                            None,
                            None,
                        )
                    else:
                        time.sleep(1)
                    continue
                info = data_return.get("INFO")
                stdout = data_return.get("STDOUT")
                stderr = data_return.get("STDERR")
                job_state = data_return.get("PROCESSING", "unknown")
                job_state = job_state.encode()
                if job_state == self.driver.job_processing:
                    time.sleep(job_processing_interval)
                    processing_attempts += 1
                    if processing_attempts > 20:
                        job_processing_interval = 1
                elif job_state == self.driver.job_failed:
                    state_timeout = time.time()
                    return (
                        False,
                        "Job Failed: {}".format(job_id),
                        stdout,
                        stderr,
                        info,
                    )
                elif job_state in [
                    self.driver.job_end,
                    self.driver.nullbyte,
                    self.driver.transfer_end,
                ]:
                    nodes = len(data_return.get("_nodes"))
                    if len(data_return.get("FAILED", list())) > 0:
                        state_timeout = time.time()
                        return (
                            False,
                            "Job Degrated: {}".format(job_id),
                            stdout,
                            stderr,
                            info,
                        )
                    elif len(data_return.get("SUCCESS", list())) == nodes:
                        state_timeout = time.time()
                        return (
                            True,
                            "Job Success: {}".format(job_id),
                            stdout,
                            stderr,
                            info,
                        )
                    else:
                        if time.time() - state_timeout > timeout:
                            state_timeout = time.time()
                            return (
                                True,
                                "Job Skipped: {}".format(job_id),
                                stdout,
                                stderr,
                                info,
                            )
                        else:
                            time.sleep(1)
                else:
                    if time.time() - state_timeout > timeout:
                        state_timeout = time.time()
                        return (
                            None,
                            "Job in an unknown state: {}".format(job_id),
                            stdout,
                            stderr,
                            info,
                        )
                    else:
                        time.sleep(1)

    def analyze_job(self, job_id):
        data = directord.send_data(
            socket_path=self.args.socket_path,
            data=json.dumps(dict(manage={"job_info": job_id})),
        )

        item = list(dict(json.loads(data.decode())).values())

        if item and not item[0]:
            return json.dumps({"job_id_not_found": job_id})

        return self.analyze_data(parent_id=job_id, parent_jobs=item)

    def analyze_parent(self, parent_id):
        data = directord.send_data(
            socket_path=self.args.socket_path,
            data=json.dumps(dict(manage={"list_jobs": None})),
        )
        parent_jobs = list()
        if data:
            data = dict(json.loads(data.decode()))
            for value in data.values():
                if value["PARENT_JOB_ID"] == parent_id:
                    parent_jobs.append(value)

        if not parent_jobs:
            return json.dumps({"parent_id_not_found": parent_id})

        return self.analyze_data(parent_id=parent_id, parent_jobs=parent_jobs)

    def analyze_data(self, parent_id, parent_jobs):
        meta = dict(
            execution=collections.defaultdict(int),
            roundtrip=collections.defaultdict(int),
            nodes=set(),
            node_successes=collections.defaultdict(int),
            node_failures=collections.defaultdict(int),
        )
        analysis = dict(id=parent_id, total_jobs=len(parent_jobs))
        for job in parent_jobs:
            for k, v in job.get("_executiontime", dict()).items():
                meta["nodes"].add(k)
                meta["execution"][k] += v

            for k, v in job.get("_roundtripltime", dict()).items():
                meta["nodes"].add(k)
                meta["roundtrip"][k] += v

            for item in job.get("SUCCESS", list()):
                meta["node_successes"][item] += 1

            for item in job.get("FAILED", list()):
                meta["node_failures"][item] += 1

        analysis["actual_runtime"] = parent_jobs[-1].get(
            "_lasttime", 0
        ) - parent_jobs[0].get("_createtime", 0)
        analysis["slowest_node_execution"] = max(
            meta["execution"], key=meta["execution"].get
        )
        analysis["slowest_node_roundtrip"] = max(
            meta["roundtrip"], key=meta["roundtrip"].get
        )
        analysis["fastest_node_execution"] = min(
            meta["execution"], key=meta["execution"].get
        )
        analysis["fastest_node_roundtrip"] = min(
            meta["roundtrip"], key=meta["roundtrip"].get
        )
        analysis["combined_execution_time"] = sum(meta["execution"].values())
        analysis["total_successes"] = sum(meta["node_successes"].values())
        analysis["total_failures"] = sum(meta["node_failures"].values())
        analysis["total_node_count"] = len(meta["nodes"])
        analysis["total_avg_execution_time"] = (
            analysis["combined_execution_time"] / analysis["total_jobs"]
        )

        return json.dumps(analysis, sort_keys=True)

    def run(self, override=None):
        """Send the management command to the server.

        :param override: Set the job function regardless of args.
        :type override: String
        :returns: String
        """

        def _cache_dump():
            with diskcache.Cache(
                self.args.cache_path,
                tag_index=True,
                disk=diskcache.JSONDisk,
            ) as cache:
                cache_dict = {}
                for item in cache.iterkeys():
                    cache_dict[item] = cache.get(item)
                print(json.dumps(cache_dict, indent=4))

        execution_map = {
            "dump-cache": _cache_dump,
            "export-jobs": {"list_jobs": None},
            "export-nodes": {"list_nodes": None},
            "generate-keys": self.generate_certificates,
            "job-info": {"job_info": override},
            "list-jobs": {"list_jobs": None},
            "list-nodes": {"list_nodes": None},
            "purge-jobs": {"purge_jobs": None},
            "purge-nodes": {"purge_nodes": None},
            "analyze-parent": self.analyze_parent,
            "analyze-job": self.analyze_job,
        }

        if override and override in execution_map:
            manage = execution_map[override]
            if callable(manage):
                return manage()
        elif isinstance(override, str):
            manage = execution_map["job-info"]
        else:
            for k, v in execution_map.items():
                k_obj = k.replace("-", "_")
                k_arg = getattr(self.args, k_obj, False)
                if k_arg:
                    if callable(v):
                        if isinstance(override, str):
                            return v(override)
                        elif isinstance(k_arg, str):
                            return v(k_arg)
                        else:
                            return v()
                    else:
                        if isinstance(k_arg, str):
                            v[k_obj] = k_arg
                        manage = v
                        break
            else:
                raise SystemExit("No known management function was defined.")

        self.log.debug("Executing Management Command:%s", manage)
        return directord.send_data(
            socket_path=self.args.socket_path,
            data=json.dumps(dict(manage=manage)),
        )



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import getpass
import os
import sys

import jinja2
from jinja2 import StrictUndefined

import yaml

import directord

from directord import logger
from directord import utils


class PrintError:
    def __init__(self) -> None:
        self.line_break = "="
        self.line_multiplier = 20
        self.start = "Start Error Information"
        self.start_length = len(self.start)

    def __enter__(self):
        line = self.line_break * self.line_multiplier
        print("{line} {start} {line}".format(line=line, start=self.start))

    def __exit__(self, *args, **kwargs):
        end = "End Error Information"
        line_delta = self.start_length - len(end)
        line = self.line_break * (self.line_multiplier + line_delta // 2)
        print("{line} {end} {line}".format(line=line, end=end))


class Bootstrap(directord.Processor):
    """Mixin class."""

    def __init__(self, args):
        """Initialize the Directord mixin.

        Sets up the mixin object.

        :param args: Arguments parsed by argparse.
        :type args: Object
        """

        super(Bootstrap, self).__init__()
        self.args = args
        self.blueprint = jinja2.Environment(
            loader=jinja2.BaseLoader(),
            keep_trailing_newline=True,
            undefined=StrictUndefined,
        )
        self.log = logger.getLogger(
            name="directord", debug_logging=getattr(args, "debug", False)
        )

    @staticmethod
    def bootstrap_catalog_entry(entry):
        """Return a flattened list of bootstrap job entries.

        :param entry: Catalog entry for bootstraping.
        :type entry: Dictionary
        :returns: List
        """

        ordered_entries = list()
        for item in ["jobs", "targets"]:
            if item not in entry:
                raise SystemExit(
                    "The bootstrap catalog is missing a"
                    " required component: {}".format(item)
                )

        args = entry.get("args", dict(port=22, username=getpass.getuser()))
        for target in entry["targets"]:
            try:
                target_host = target["host"]
            except KeyError:
                raise SystemExit("[host] is undefined in bootstrap catalog")
            item = dict(
                host=target_host,
                username=target.get(
                    "username", args.get("username", getpass.getuser())
                ),
                port=target.get("port", args.get("port", 22)),
                jobs=entry["jobs"],
            )
            ordered_entries.append(item)
        return ordered_entries

    @staticmethod
    def bootstrap_localfile_padding(localfile):
        """Return a padded localfile.

        Local files should be a fully qualified path. If the path is
        not fully qualified, this method will add the tools prefix to the
        file.

        :param localfile: Path to file.
        :type localfile: String
        :returns: String
        """

        if not localfile.startswith(os.sep):
            if sys.prefix == sys.base_prefix:
                base_path = os.path.join(
                    sys.base_prefix, "share/directord/tools"
                )
            else:
                base_path = os.path.join(sys.prefix, "share/directord/tools")
            return os.path.join(base_path, localfile)

        return localfile

    def bootstrap_flatten_jobs(self, jobs, return_jobs=None):
        """Return a flattened list of jobs.

        This method will flatten a list of jobs, and if an entry is an array
        the method will recurse.

        :param jobs: List of jobs to parse.
        :type jobs: List
        :param return_jobs: Seed list to use when flattening the jobs array.
        :type: return_jobs: None|List
        :returns: List
        """

        if not return_jobs:
            return_jobs = list()

        for job in jobs:
            if isinstance(job, list):
                return_jobs = self.bootstrap_flatten_jobs(
                    jobs=job, return_jobs=return_jobs
                )
            else:
                return_jobs.append(job)

        return return_jobs

    @staticmethod
    def _read_chunks(fh, chunk_size=131072):
        """Read file in 2048 chunks."""

        while True:
            data = fh.read(chunk_size)
            if not data:
                break
            yield data

    def bootstrap_file_send(self, ssh, localfile, remotefile):
        """Run a remote put command.

        :param ssh: SSH connection object.
        :type ssh: Object
        :param localfile: Local file to transfer.
        :type localfile: String
        :param remotefile: Remote file destination.
        :type remotefile: String
        """

        if "sftp" in ssh.channels:
            chan = ssh.channels["sftp"]
        else:
            chan = ssh.channels["sftp"] = ssh.session.sftp_new()
        if isinstance(chan, int):
            with PrintError():
                self.log.critical(
                    "Failed to connect for file transfer ADD. Error Code"
                    " [ %s ]",
                    chan,
                )
                raise SystemExit(chan)

        chan.init()
        self.log.debug(
            "channel: %s, local file: %s, remote file: %s",
            chan,
            localfile,
            remotefile,
        )
        fileinfo = os.stat(localfile)
        try:
            try:
                chan.unlink(remotefile)
            except Exception as e:
                self.log.debug("file unlink error %s", str(e))

            flags = os.O_WRONLY | os.O_CREAT | os.O_APPEND
            with open(localfile, "rb") as local_f:
                for data in self._read_chunks(fh=local_f, chunk_size=131072):
                    with chan.open(
                        remotefile, flags, fileinfo.st_mode
                    ) as remote_f:
                        remote_f.write(data)
        except Exception as e:
            with PrintError():
                self.log.critical(str(e))
                raise SystemExit("File [ {} ] ADD failed.".format(remotefile))
        else:
            self.log.info(
                "HOST: [ %s ] SUCCESS: ADD [ %s ] TO [ %s ]",
                ssh.host,
                localfile,
                remotefile,
            )

    def bootstrap_file_get(self, ssh, localfile, remotefile):
        """Run a remote get command.

        :param ssh: SSH connection object.
        :type ssh: Object
        :param localfile: Local file destination.
        :type localfile: String
        :param remotefile: Remote file to transfer.
        :type remotefile: String
        """

        if "sftp" in ssh.channels:
            chan = ssh.channels["sftp"]
        else:
            chan = ssh.channels["sftp"] = ssh.session.sftp_new()
        if isinstance(chan, int):
            with PrintError():
                self.log.critical(
                    "Failed to connect for file transfer GET. Error Code"
                    " [ %s ]",
                    chan,
                )
                raise SystemExit(chan)

        chan.init()
        self.log.debug(
            "channel: %s, local file: %s, remote file: %s",
            chan,
            localfile,
            remotefile,
        )
        try:
            with chan.open(remotefile, os.O_RDONLY, 0) as remote_f:
                with open(localfile, "wb") as f:
                    for _, data in remote_f:
                        f.write(data)
        except Exception as e:
            with PrintError():
                self.log.critical(str(e))
                raise SystemExit("File [ {} ] GET failed.".format(remotefile))
        else:
            self.log.info(
                "HOST: [ %s ] SUCCESS: GET [ %s ] TO [ %s ]",
                ssh.host,
                remotefile,
                localfile,
            )

    def bootstrap_exec(self, ssh, command, catalog):
        """Run a remote command.

        Run a command and check the status. If there's a failure the
        method will exit error.

        :param session: SSH Session connection object.
        :type session: Object
        :param command: Plain-text execution string.
        :type command: String
        :param catalog: The job catalog definition.
        :type catalog: Dictionary
        """

        chan = ssh.session.channel_new()
        if isinstance(chan, int):
            self.log.critical(
                "Failed to connect for remote execution."
                " SSH Error Code [ %s ]",
                chan,
            )
            raise SystemExit(chan)

        try:
            chan.open_session()
            if chan.request_pty() == 0:
                self.log.debug("using PTY")
            t_command = self.blueprint.from_string(command)
            command = t_command.render(**catalog)
            self.log.debug(
                "channel: %s, command: %s",
                chan,
                command,
            )
            chan.request_exec(command)
            size, data = chan.read()
            while size > 0:
                size, _data = chan.read()
                data += _data
            try:
                data = data.decode()
            except AttributeError:
                pass

            if chan.get_exit_status() != 0:
                self.log.warning("FAILURE: [ %s ]", command)
                with PrintError():
                    self.log.critical(data)
                    raise SystemExit(
                        "Bootstrap command failed: {}".format(command)
                    )
            else:
                self.log.info(
                    "HOST: [ %s ] SUCCESS: [ %s ]", ssh.host, command
                )
        finally:
            if hasattr(chan, "close"):
                chan.close()

    def bootstrap_run(self, job_def, catalog):
        """Run a given set of jobs using a defined job definition.

        This method requires a job definition which contains the following.

        {
            "host": String,
            "port": Integer,
            "username": String,
            "key_file": String,
            "jobs": List,
        }

        :param jobs_def: Defined job definition.
        :type jobs_def: Dictionary
        :param catalog: The job catalog definition.
        :type catalog: Dictionary
        """

        self.log.info("Running bootstrap for %s", job_def["host"])
        with utils.SSHConnect(
            host=job_def["host"],
            username=job_def["username"],
            port=job_def["port"],
            key_file=job_def.get("key_file"),
            debug=getattr(self.args, "debug", False),
        ) as ssh:
            for job in self.bootstrap_flatten_jobs(jobs=job_def["jobs"]):
                key, value = next(iter(job.items()))
                self.log.debug("Executing: %s %s", key, value)
                with self.timeout(
                    time=600, job_id=job_def["host"], reraise=True
                ):
                    if key == "RUN":
                        self.bootstrap_exec(
                            ssh=ssh, command=value, catalog=catalog
                        )
                    elif key == "ADD":
                        localfile, remotefile = value.split(" ", 1)
                        localfile = self.bootstrap_localfile_padding(localfile)
                        self.bootstrap_file_send(
                            ssh=ssh, localfile=localfile, remotefile=remotefile
                        )
                    elif key == "GET":
                        remotefile, localfile = value.split(" ", 1)
                        self.bootstrap_file_get(
                            ssh=ssh, localfile=localfile, remotefile=remotefile
                        )

    def bootstrap_q_processor(self, queue, catalog):
        """Run a queing execution thread.

        The queue will be processed so long as there are objects to process.

        :param queue: SSH connection object.
        :type queue: Object
        :param catalog: The job catalog definition.
        :type catalog: Dictionary
        """

        while True:
            try:
                job_def = queue.get(timeout=3)
            except Exception:
                break
            else:
                self.bootstrap_run(
                    job_def=job_def,
                    catalog=catalog,
                )

    def bootstrap_cluster(self):
        """Run a cluster wide bootstrap using a catalog file.

        Cluster bootstrap requires a catalog file to run. Catalogs are broken
        up into two sections, `directord_server` and `directord_client`. All
        servers are processed serially and first. All clients are processing
        in parallel using a maximum of the threads argument.
        """

        q = self.get_queue()
        catalog = dict()
        if not self.args.catalog:
            raise SystemExit("No catalog was defined.")

        for c in self.args.catalog:
            utils.merge_dict(base=catalog, new=yaml.safe_load(c))

        directord_server = catalog.get("directord_server")
        if directord_server:
            self.log.info("Loading server information")
            for s in self.bootstrap_catalog_entry(entry=directord_server):
                s["key_file"] = self.args.key_file
                self.bootstrap_run(job_def=s, catalog=catalog)

        directord_clients = catalog.get("directord_clients")
        if directord_clients:
            self.log.info("Loading client information")
            for c in self.bootstrap_catalog_entry(entry=directord_clients):
                c["key_file"] = self.args.key_file
                q.put(c)

        threads = list()
        for _ in range(self.args.threads):
            threads.append(
                (
                    self.thread(
                        target=self.bootstrap_q_processor, args=(q, catalog)
                    ),
                    True,
                )
            )
        else:
            self.run_threads(threads=threads)



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import grp
import os
import pwd
import traceback

from directord import components


class Component(components.ComponentBase):
    def __init__(self):
        """Initialize the component cache class."""

        super().__init__(desc="Process workdir commands")

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "--chown", help="Set the file ownership", type=str
        )
        self.parser.add_argument("--chmod", help="Set the file mode", type=str)
        self.parser.add_argument("workdir", help="Create a directory.")

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        if self.known_args.chown:
            chown = self.known_args.chown.split(":", 1)
            if len(chown) == 1:
                chown.append(None)
            data["user"], data["group"] = chown

        if self.known_args.chmod:
            data["mode"] = int(oct(int(self.known_args.chmod, 8)), 8)

        data["workdir"] = self.known_args.workdir
        return data

    def client(self, cache, job):
        """Run file work directory operation.

        Command operations are rendered with cached data from the args dict.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        self.log.debug("client(): job: %s, cache: %s", job, cache)
        success, workdir = self.blueprinter(
            content=job["workdir"],
            values=cache.get("args"),
            allow_empty_values=True,
        )
        if not success:
            return None, workdir, False, None

        user = job.get("user")
        group = job.get("group")
        mode = job.get("mode")

        if not workdir:
            return None, None, False, None

        try:
            os.makedirs(workdir, exist_ok=True)
        except (FileExistsError, PermissionError) as e:
            self.log.critical(str(e))
            return None, traceback.format_exc(), False
        else:
            update_info = "Directory {} OK".format(workdir)
            outcome = True
            stderr = None
            info = ""
            if user:
                try:
                    try:
                        uid = int(user)
                    except ValueError:
                        uid = pwd.getpwnam(user).pw_uid
                        info += " uid {} found from name {}.".format(uid, user)

                    if group:
                        try:
                            gid = int(group)
                        except ValueError:
                            gid = grp.getgrnam(group).gr_gid
                            info += " gid {} found from name {}.".format(
                                gid, group
                            )
                    else:
                        gid = -1
                except KeyError as e:
                    outcome = False
                    stderr = (
                        "Failed to set ownership properties."
                        " USER:{} GROUP:{}".format(user, group)
                    )
                    info += " {}".format(str(e))
                else:
                    os.chown(workdir, uid, gid)
                    outcome = True

            if mode:
                os.chmod(workdir, mode)

            return update_info, stderr, outcome, info



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import glob
import grp
import os
import pwd
import shlex
import time
import traceback

from directord import components
from directord import utils


class Transfer:
    """Transfer connection context manager."""

    def __init__(self, driver, log, job_id):
        """Initialize the transfer context manager class."""

        self.driver = driver
        self.bind_transfer = self.driver.backend_connect()
        self.log = log
        self.job_id = job_id

    def __enter__(self):
        """Return the bind_transfer object on enter."""

        self.log.debug(
            "Transfer started to %s for job [ %s ]",
            self.driver.identity,
            self.job_id,
        )
        return self.bind_transfer

    def __exit__(self, *args, **kwargs):
        """Close the bind transfer object."""

        try:
            self.bind_transfer.close(linger=2)
            close_time = time.time()
            while not self.bind_transfer.closed:
                if time.time() - close_time > 60:
                    raise TimeoutError(
                        "Job [ {} ] failed to close transfer socket".format(
                            self.job_id
                        )
                    )
                else:
                    self.bind_transfer.close(linger=2)
                    self.delay(1)
        except Exception as e:
            self.log.error(
                "Job [ %s ] transfer ran into an exception"
                " while closing the socket %s",
                str(e),
                self.job_id,
            )
        else:
            self.log.debug("Job [ %s ] transfer socket closed", self.job_id)

        self.log.debug(
            "Transfer ended for %s for job [ %s ]",
            self.driver.identity,
            self.job_id,
        )
        time.sleep(0.25)


class Component(components.ComponentBase):
    def __init__(self):
        """Initialize the component cache class."""

        super().__init__(desc="Process transfer commands")
        self.requires_lock = True

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "--chown", help="Set the file ownership", type=str
        )
        self.parser.add_argument("--chmod", help="Set the file mode", type=str)
        self.parser.add_argument(
            "--blueprint",
            action="store_true",
            help="Instruct the remote file to be blueprinted.",
        )
        self.parser.add_argument(
            "files",
            nargs="+",
            help="Set the file to transfer: 'FROM' 'TO'",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        if self.known_args.chown:
            chown = self.known_args.chown.split(":", 1)
            if len(chown) == 1:
                chown.append(None)
            data["user"], data["group"] = chown

        if self.known_args.chmod:
            data["mode"] = int(oct(int(self.known_args.chmod, 8)), 8)

        file_from, data["to"] = shlex.split(" ".join(self.known_args.files))
        data["from"] = [
            os.path.abspath(os.path.expanduser(i))
            for i in glob.glob(file_from)
            if os.path.isfile(os.path.expanduser(i))
        ]
        if not data["from"]:
            raise AttributeError(
                "The value of [ {} ] was not found.".format(file_from)
            )
        data["blueprint"] = self.known_args.blueprint

        return data

    def client(self, cache, job):
        """Run file transfer operation.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        driver = self.driver.__copy__()
        self.log.debug("client(): job: %s, cache: %s", job, cache)
        with Transfer(
            driver=driver, log=self.log, job_id=job["job_id"]
        ) as bind_transfer:
            return self._client(cache, job, self.info, driver, bind_transfer)

    def _client(self, cache, job, source_file, driver, bind_transfer):
        """Run file transfer operation.

        File transfer operations will look at the cache, then look for an
        existing file, and finally compare the original SHA3_224 to what is on
        disk. If everything checks out the client will request the file
        from the server.

        If the user and group arguments are defined the file ownership
        will be set accordingly.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :param source_file: Original file location on server.
        :type source_file: String
        :param driver: Connection object used to store information used in a
                     return message.
        :type driver: Object
        :returns: tuple
        """

        file_to = job["file_to"]
        user = job.get("user")
        group = job.get("group")
        file_sha3_224 = job.get("file_sha3_224")
        blueprint = job.get("blueprint", False)
        success, file_to = self.blueprinter(
            content=file_to, values=cache.get("args"), allow_empty_values=True
        )
        if not success:
            return None, file_to, False, None

        mode = job.get("mode")
        if (
            os.path.isfile(file_to)
            and utils.file_sha3_224(file_to) == file_sha3_224
        ):
            stdout = (
                "File exists {} and SHA3_224 {} matches, nothing to"
                " transfer".format(file_to, file_sha3_224)
            )
            if blueprint:
                success, error = self.file_blueprinter(
                    cache=cache, file_to=file_to
                )
                if not success:
                    return utils.file_sha3_224(file_to), error, False, None

            return stdout, None, True, None
        else:
            self.log.debug(
                "Job [ %s ] requesting transfer of source file:%s",
                job["job_id"],
                source_file,
            )

        try:
            offset = 0
            chunk = 131072
            with open(file_to, "wb") as f:
                while True:
                    driver.socket_send(
                        socket=bind_transfer,
                        msg_id=job["job_id"].encode(),
                        control=driver.transfer_start,
                        command="{}".format(offset).encode(),
                        data="{}".format(chunk).encode(),
                        info=source_file,
                    )
                    offset += chunk
                    (
                        _,
                        control,
                        _,
                        data,
                        info,
                        _,
                        _,
                    ) = driver.socket_recv(socket=bind_transfer)
                    if control in [driver.job_processing, driver.transfer_end]:
                        chunk_size = len(data)
                        self.log.debug(
                            "Job [ %s ] identity [ %s ] received %s",
                            job["job_id"],
                            driver.identity,
                            chunk_size,
                        )
                        f.write(data)
                        if control == driver.transfer_end:
                            self.log.debug(
                                "Job [ %s ] identity [ %s ] stopped transfer",
                                job["job_id"],
                                driver.identity,
                            )
                            break
                        elif chunk_size < chunk:
                            self.log.debug(
                                "Job [ %s ] identity [ %s ] received the last"
                                " chunk",
                                job["job_id"],
                                driver.identity,
                            )
                            break
                    elif control == driver.job_failed:
                        return (
                            None,
                            "Transfer failed: {}".format(info.decode),
                            False,
                            None,
                        )
        except (FileNotFoundError, NotADirectoryError) as e:
            self.log.critical(
                "Job [ %s ] file failure: %s", job["job_id"], str(e)
            )
            return None, traceback.format_exc(), False, None
        except Exception as e:
            return (
                None,
                "Transfer never started: {}".format(str(e)),
                False,
                None,
            )
        else:
            self.log.debug(
                "Job [ %s ] transfer of source file:%s complete to [ %s ]",
                job["job_id"],
                source_file,
                file_to,
            )

        if blueprint:
            success, error = self.file_blueprinter(
                cache=cache, file_to=file_to
            )
            if not success:
                return utils.file_sha3_224(file_to), error, False, None

        stderr = None
        outcome = True
        if user:
            try:
                try:
                    uid = int(user)
                except ValueError:
                    uid = pwd.getpwnam(user).pw_uid

                if group:
                    try:
                        gid = int(group)
                    except ValueError:
                        gid = grp.getgrnam(group).gr_gid
                else:
                    gid = -1
            except KeyError:
                outcome = False
                stderr = (
                    "Failed to set ownership properties."
                    " USER:{} GROUP:{}".format(user, group)
                )
            else:
                os.chown(file_to, uid, gid)
                outcome = True

        if mode:
            os.chmod(file_to, mode)

        if not blueprint and utils.file_sha3_224(file_to) != file_sha3_224:
            stderr = (
                "Data integrity failure. Expected SHA {}, found SHA {}."
                " Check transfer logs for more details.".format(
                    utils.file_sha3_224(file_to), file_sha3_224
                )
            )
            return None, stderr, False, None

        return utils.file_sha3_224(file_to), stderr, outcome, None



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import time

import diskcache


class BaseDocument:
    """Create a document store object."""

    def __init__(self, url):
        """Initialize the redis datastore.

        :param url: Connection string to the redis deployment.
        :type url: String
        """

        self.datastore = diskcache.Cache(
            url,
            tag_index=True,
            disk=diskcache.JSONDisk,
        )

    def __getitem__(self, key):
        """Return the value of a given key.

        Objects stored in the keyspace are JSON, the getter will
        decode and load the JSON into a standard python object.

        :param key: Named object.
        :type key: Object
        :returns: Object
        """

        return self.datastore.get(key)

    def __setitem__(self, key, value):
        """Set an item in the datastore.

        objects are serialized JSON.

        :param key: Named object to set.
        :type key: Object
        :param value: Object to set.
        :type value: Object
        """

        try:
            key = key.decode()
        except AttributeError:
            pass

        if isinstance(value, dict):
            try:
                expire = int(value.get("time") - time.time())
            except TypeError:
                expire = None
            else:
                if expire < 1:
                    expire = 1
        else:
            expire = None

        self.datastore.set(key, value, expire=expire)

    def __delitem__(self, key):
        """Delete an item from the datastore.

        :param key: Named object.
        :type key: Object
        """

        self.datastore.pop(key)

    def items(self):
        """Yield a tuple for key and value."""

        for item in list(self.datastore):
            yield item, self.__getitem__(item)

    def keys(self):
        """Return an array of all keys.

        :returns: List
        """

        return [i.encode() for i in self.datastore]

    def empty(self):
        """Empty all items from the datastore.

        Because a Manager Dict is a proxy object we don't want to replace the
        object we want to empty it keeping the original proxy intact. This
        method will pop all items from the object.
        """

        self.datastore.clear(retry=True)

    def pop(self, key):
        """Delete the value of a given key.

        :param key: Named object.
        :type key: Object
        """

        self.__delitem__(key)

    def prune(self):
        """Prune items that have a time based expiry."""

        self.datastore.expire()
        return len(self.keys())

    def get(self, key):
        """Return the value of a given key.

        :param key: Named object.
        :type key: Object
        :returns: Object
        """

        return self.__getitem__(key)

    def set(self, key, value):
        """Set key and value if key doesn't already exist.

        :param key: Named object to set.
        :type key: Object
        :param value: Object to set.
        :type value: Object
        :param value: Object to set.
        :type value: Object
        :returns: Object
        """

        item = self.__getitem__(key)
        if item:
            return item

        self.__setitem__(key, value)
        return value



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import json
import os
import traceback

import yaml

try:
    from podman import PodmanClient

    AVAILABLE_PODMAN = True
except ImportError:
    AVAILABLE_PODMAN = False

from directord import components


class Component(components.ComponentBase):
    def __init__(self):
        """Initialize the component cache class."""

        super().__init__(desc="Process pod commands")

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "--socket-path",
            default="/var/run/podman/podman.sock",
            help="Path to the podman socket. Default: %(default)s",
        )
        self.parser.add_argument(
            "--env",
            help="Comma separated environment variables. KEY=VALUE,...",
            metavar="KEY=VALUE",
        )
        self.parser.add_argument(
            "--command",
            help="Run a command in an exec container.",
            nargs="+",
        )
        self.parser.add_argument(
            "--privileged",
            action="store_true",
            help="Access a container with privleges.",
        )
        self.parser.add_argument(
            "--tls-verify",
            action="store_true",
            help="Verify certificates when pulling container images.",
        )
        self.parser.add_argument(
            "--force",
            action="store_true",
            help="When running removal operations, Enable|Disable force.",
        )
        self.parser.add_argument(
            "--kill-signal",
            default="SIGKILL",
            help="Set the kill signal. Default: %(default)s",
            metavar="SIGNAL",
        )
        pod_group = self.parser.add_mutually_exclusive_group(required=True)
        pod_group.add_argument(
            "--start", help="Start a pod.", metavar="POD_NAME"
        )
        pod_group.add_argument(
            "--stop", help="Stop a pod.", metavar="POD_NAME"
        )
        pod_group.add_argument(
            "--rm", help="Remove a pod.", metavar="POD_NAME"
        )
        pod_group.add_argument(
            "--kill", help="Kill a pod.", metavar="POD_NAME"
        )
        pod_group.add_argument(
            "--inspect", help="Inspect a pod.", metavar="POD_NAME"
        )
        pod_group.add_argument(
            "--play",
            help="Play a pod from a structured file.",
            metavar="POD_FILE",
        )
        pod_group.add_argument(
            "--exec-run",
            help="Create an execution container to run a command within.",
            metavar="CONTAINER_NAME",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        if self.known_args.start:
            data["pod_action"] = "start"
            data["kwargs"] = dict(
                name=self.known_args.start, timeout=self.known_args.timeout
            )
        elif self.known_args.stop:
            data["pod_action"] = "stop"
            data["kwargs"] = dict(
                name=self.known_args.stop, timeout=self.known_args.timeout
            )
        elif self.known_args.rm:
            data["pod_action"] = "rm"
            data["kwargs"] = dict(
                name=self.known_args.rm, force=self.known_args.force
            )
        elif self.known_args.kill:
            data["pod_action"] = "kill"
            data["kwargs"] = dict(
                name=self.known_args.kill, signal=self.known_args.kill_signal
            )
        elif self.known_args.inspect:
            data["pod_action"] = "inspect"
            data["kwargs"] = dict(name=self.known_args.inspect)
        elif self.known_args.play:
            data["pod_action"] = "play"
            data["kwargs"] = dict(
                pod_file=self.known_args.play,
                tls_verify=self.known_args.tls_verify,
            )
        elif self.known_args.exec_run:
            data["pod_action"] = "exec_run"
            data["kwargs"] = dict(
                name=self.known_args.exec_run,
                privileged=self.known_args.privileged,
                command=self.known_args.command,
            )
            if self.known_args.env:
                data["kwargs"]["env"] = self.known_args.env.split(",")

        data["socket_path"] = self.known_args.socket_path
        return data

    def client(self, cache, job):
        """Run pod command operation.

        :param command: Work directory path.
        :type command: String
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        if not AVAILABLE_PODMAN:
            return (
                None,
                "The required podman-py library is not installed",
                False,
                None,
            )
        try:
            with PodmanPod(socket=job["socket_path"]) as p:
                action = getattr(p, job["pod_action"], None)
                if action:
                    status, data = action(**job["kwargs"])
                    if data:
                        data = json.dumps(data)
                    if status:
                        return data, None, status, None

                    return None, data, status, None

                return (
                    None,
                    (
                        "The action [ {action} ] failed to return"
                        "  a function".format(action=job["pod_action"])
                    ),
                    False,
                    None,
                )
        except Exception as e:
            self.log.critical(
                "Job [ %s ] critical error %s", job["job_id"], str(e)
            )
            return None, traceback.format_exc(), False, None


class PodmanConnect:
    """Connect to the podman unix socket."""

    def __init__(self, socket="/var/run/podman/podman.sock"):
        """Initialize the Directord pod connection class.

        Sets up the pod api object.

        :param socket: Socket path to connect to.
        :type socket: String
        """

        self.pod = PodmanClient(
            base_url="unix://{socket}".format(socket=socket)
        )
        self.api = self.pod.api

    def __exit__(self, *args, **kwargs):
        """Close the connection and exit."""

        self.close()

    def __enter__(self):
        """Return self for use within a context manager."""

        return self

    def close(self):
        """Close the connection."""

        self.pod.close()


class PodmanPod(PodmanConnect):
    def __init__(self, socket):
        """Initialize the Directord pod class.

        Sets up the pod api object.

        :param socket: Socket path to connect to.
        :type socket: String
        """

        super(PodmanPod, self).__init__(socket=socket)

    @staticmethod
    def _decode(data):
        """Decode byte data and return it.

        If the byte data is JSON type the returned value will
        be a JSON string.
        """

        data = data.decode()
        if data:
            try:
                return json.loads(data)
            except Exception:
                return data

    def start(self, name, timeout=120):
        """Start a given pod and return the action status.

        :param name: Pod name.
        :type name: String
        :param timeout: API timeout when running a pod interaction.
        :type timeout: Integer
        :returns: Tuple
        """

        resp = self.api.post(
            path="/pods/{name}/start".format(name=name),
            params={"t": timeout},
        )
        return resp.ok, self._decode(resp.content)

    def stop(self, name, timeout=120):
        """Stop a given pod and return the action status.

        :param name: Pod name.
        :type name: String
        :param timeout: API timeout when running a pod interaction.
        :type timeout: Integer
        :returns: Tuple
        """

        resp = self.api.post(
            path="/pods/{name}/stop".format(name=name),
            params={"t": timeout},
        )
        return resp.ok, self._decode(resp.content)

    def rm(self, name, force=False):
        """Remove a given pod and return the action status.

        :param name: Pod name.
        :type name: String
        :param force: Force remove a pod
        :type force: Boolean
        :returns: Tuple
        """

        resp = self.api.delete(
            path="/pods/{name}".format(name=name),
            params={"force": force},
        )
        return resp.ok, self._decode(resp.content)

    def kill(self, name, signal="SIGKILL"):
        """Kill a given pod and return the action status.

        :param name: Pod name.
        :type name: String
        :param signal: Signal used when killing a pod
        :type signal: String
        :returns: Tuple
        """

        resp = self.api.post(
            path="/pods/{name}/kill".format(name=name),
            params={"signal": signal},
        )
        return resp.ok, self._decode(resp.content)

    def inspect(self, name):
        """Inspect a given pod and return the action status.

        :param name: Pod name.
        :type name: String
        :returns: Tuple
        """

        resp = self.api.get(path="/pods/{name}/json".format(name=name))
        return resp.ok, self._decode(resp.content)

    def play(self, pod_file, tls_verify=True):
        """Instantiate a given pod and return the action status.

        :param pod_file: Full path to a pod YAML file.
        :type pod_file: String
        :returns: Tuple
        """

        if os.path.exists(os.path.abspath(os.path.expanduser(pod_file))):
            with open(pod_file) as f:
                resp = self.api.post(
                    path="/play/kube",
                    data=json.dumps(yaml.safe_load(f)),
                    params={"tlsVerify": tls_verify, "start": True},
                )
            return resp.ok, self._decode(resp.content)

        return False, "Pod YAML did not exist"

    def exec_run(self, name, command, env, privileged=False):
        """Create an exec container and run it.

        :param name: Container name.
        :type name: String
        :param command: Command to be run in list format.
        :type command: List
        :param env: Environment used when executing a command.
        :type env: List
        :param privileged: Enable|Disable privileged commands.
        :type privileged: Boolean
        :returns: Tuple
        """

        _command = {
            "AttachStderr": True,
            "AttachStdin": True,
            "AttachStdout": True,
            "Cmd": command,
            "Env": env,
            "Privileged": privileged,
            "Tty": True,
        }
        resp = self.api.post(
            path="containers/{name}/exec".format(name=name),
            data=json.dumps(_command),
        )
        if resp.ok:
            resp = self.api.post(
                path="/exec/{id}/start".format(
                    id=self._decode(resp.content)["Id"]
                ),
                data=json.dumps({"Detach": False, "Tty": True}),
            )
        return resp.ok, self._decode(resp.content)



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from directord import components
from directord import utils


class Component(components.ComponentBase):
    def __init__(self):
        """Initialize the component cache class."""

        super().__init__(desc="Process run commands")

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "--stdout-arg",
            help="Stores the stdout of a given command as a cached argument.",
        )
        self.parser.add_argument(
            "--no-block",
            action="store_true",
            help="Run a command in 'fire and forget' mode.",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        if self.known_args.stdout_arg:
            data["stdout_arg"] = self.known_args.stdout_arg
        data["no_block"] = self.known_args.no_block
        data["command"] = " ".join(self.unknown_args)

        return data

    def client(self, cache, job):
        """Run file command operation.

        Command operations are rendered with cached data from the args dict.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        self.log.debug("client(): job: %s, cache: %s", job, cache)
        stdout_arg = job.get("stdout_arg")
        success, command = self.blueprinter(
            content=job["command"],
            values=cache.get("args"),
            allow_empty_values=True,
        )
        if not success:
            return None, command, False, None
        elif not command:
            return None, None, False, None

        stdout, stderr, outcome = self.run_command(
            command=command,
            env=cache.get("envs"),
            no_block=job.get("no_block"),
        )

        if stdout_arg and stdout:
            self.block_on_tasks = list()
            clean_info = stdout.decode().strip()
            arg_job = job.copy()
            arg_job.pop("parent_sha3_224", None)
            arg_job.pop("parent_id", None)
            arg_job.pop("job_sha3_224", None)
            arg_job.pop("job_id", None)
            arg_job["skip_cache"] = True
            arg_job["extend_args"] = True
            arg_job["verb"] = "ARG"
            arg_job["args"] = {stdout_arg: clean_info}
            arg_job["parent_async_bypass"] = True
            arg_job["targets"] = [self.driver.identity]
            arg_job["job_id"] = utils.get_uuid()
            arg_job["job_sha3_224"] = utils.object_sha3_224(obj=arg_job)
            arg_job["parent_id"] = utils.get_uuid()
            arg_job["parent_sha3_224"] = utils.object_sha3_224(obj=arg_job)
            self.block_on_tasks.append(arg_job)
        elif stdout_arg and not stdout:
            return stdout, stderr, False, command.encode()

        return stdout, stderr, outcome, command.encode()



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import time


class BaseDocument(dict):
    """Create a document store object."""

    def empty(self):
        """Empty all items from the datastore.

        Because a Manager Dict is a proxy object we don't want to replace the
        object we want to empty it keeping the original proxy intact. This
        method will pop all items from the object.
        """

        try:
            while self.popitem():
                pass
        except KeyError:
            pass

    def prune(self):
        """Prune items that have a time based expiry."""

        for (key, value) in list(self.items()):
            try:
                if time.time() >= value["time"]:
                    self.pop(key)
            except (KeyError, TypeError):
                pass

        return len(self)

    def set(self, key, value):
        """Set key and value if key doesn't already exist.

        :param key: Named object to set.
        :type key: Object
        :param value: Object to set.
        :type value: Object
        :returns: Object
        """

        if key in self:
            return self[key]

        super().__setitem__(key, value)
        return self[key]

    def __repr__(self):
        return f"{type(self).__name__}({super().__repr__()})"



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import socket
import time

from directord import logger
from directord import utils


class BaseDriver:
    nullbyte = b"\000"  # Signals null
    coordination_failed = b"\007"  # Signals worker is ready
    coordination_ack = b"\020"  # Signals worker is ready
    coordination_notice = b"\021"  # Signals worker is ready

    transfer_start = b"\002"  # Signals start file transfer
    transfer_end = b"\003"  # Signals start file transfer

    heartbeat_notice = b"\005"  # Signals worker heartbeat
    job_ack = b"\006"  # Signals job started
    job_end = b"\004"  # Signals job ended
    job_failed = b"\025"  # Signals job failed
    job_processing = b"\026"  # Signals job running

    def __init__(
        self, args, encrypted_traffic_data=None, connection_string=None
    ):
        """Initialize the Driver.

        :param args: Arguments parsed by argparse.
        :type args: Object
        "param encrypted_traffic: Enable|Disable encrypted traffic.
        :type encrypted_traffic: Boolean
        """

        self.encrypted_traffic_data = encrypted_traffic_data
        self.connection_string = connection_string
        self.identity = socket.gethostname()
        self.log = logger.getLogger(name="directord")
        self.args = args

    def __copy__(self):
        """Return a copy of the base class.

        :returns: Object
        """

        return self

    def socket_send(
        self,
        socket,
        identity=None,
        msg_id=None,
        control=None,
        command=None,
        data=None,
        info=None,
        stderr=None,
        stdout=None,
    ):
        """Send a message over a socket.

        All message information is assumed to be byte encoded.

        All possible control characters are defined within the Interface class.
        For more on control characters review the following
        URL(https://donsnotes.com/tech/charsets/ascii.html#cntrl).

        :param socket: Socket object.
        :type socket: Object
        :param identity: Target where message will be sent.
        :type identity: Bytes
        :param msg_id: ID information for a given message. If no ID is
                       provided a UUID will be generated.
        :type msg_id: Bytes
        :param control: ASCII control charaters.
        :type control: Bytes
        :param command: Command definition for a given message.
        :type command: Bytes
        :param data: Encoded data that will be transmitted.
        :type data: Bytes
        :param info: Encoded information that will be transmitted.
        :type info: Bytes
        :param stderr: Encoded error information from a command.
        :type stderr: Bytes
        :param stdout: Encoded output information from a command.
        :type stdout: Bytes
        """

        if not msg_id:
            msg_id = utils.get_uuid().encode()

        if not control:
            control = self.nullbyte

        if not command:
            command = self.nullbyte

        if not data:
            data = self.nullbyte

        if not info:
            info = self.nullbyte

        if not stderr:
            stderr = self.nullbyte

        if not stdout:
            stdout = self.nullbyte

        pass

    @staticmethod
    def socket_recv(socket):
        """Receive a message over a socket.

        :param socket: socket object.
        :type socket: Object
        """

        pass

    def job_connect(self):
        """Connect to a job socket and return the socket.

        :returns: Object
        """

        pass

    def backend_connect(self):
        """Connect to a transfer socket and return the socket.

        :returns: Object
        """

        pass

    def heartbeat_connect(self):
        """Connect to a heartbeat socket and return the socket.

        :returns: Object
        """

        pass

    def heartbeat_bind(self):
        """Bind an address to a heartbeat socket and return the socket.

        :returns: Object
        """

        pass

    def heartbeat_reset(self, bind_heatbeat=None):
        """Reset the connection on the heartbeat socket.

        Returns a new ttl after reconnect.

        :param bind_heatbeat: heart beat bind object.
        :type bind_heatbeat: Object
        :returns: Tuple
        """

        return (
            self.get_heartbeat(interval=self.args.heartbeat_interval),
            self.heartbeat_connect(),
        )

    def job_bind(self):
        """Bind an address to a job socket and return the socket.

        :returns: Object
        """

        pass

    def backend_bind(self):
        """Bind an address to a transfer socket and return the socket.

        :returns: Object
        """

        pass

    def bind_check(self, bind, interval=1, constant=1000):
        """Return True if a bind type contains work ready.

        :param bind: A given Socket bind to identify.
        :type bind: Object
        :param interval: Exponential Interval used to determine the polling
                         duration for a given socket.
        :type interval: Integer
        :param constant: Constant time used to poll for new jobs.
        :type constant: Integer
        :returns: Object
        """

        pass

    def key_generate(self, keys_dir, key_type):
        """Generate certificate.

        :param keys_dir: Full Directory path where a given key will be stored.
        :type keys_dir: String
        :param key_type: Key type to be generated.
        :type key_type: String
        """

        pass

    def get_heartbeat(self, interval=0):
        """Return a new hearbeat interval time.

        :param interval: Padding for heartbeat interval.
        :type interval: Integer|Float
        :returns: Float
        """

        return time.time() + interval

    def get_expiry(self, heartbeat_interval=60, interval=3):
        """Return a new expiry time.

        :param interval: Exponential back off for expiration.
        :type interval: Integer|Float
        :returns: Float
        """

        return time.time() + (heartbeat_interval * interval)

    def create_proxy(front, back):
        """Create proxy bind.

        Bind two interfaces into a proxy.

        :param front: Frontend interface.
        :type front: Object
        :param back: Backend interface.
        :type back: Object
        """

        pass



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from multiprocessing.managers import MakeProxyType
from multiprocessing.managers import SyncManager

from directord import datastores


BaseDictProxy = MakeProxyType(
    "BaseDictProxy",
    (
        "__contains__",
        "__delitem__",
        "__getitem__",
        "__iter__",
        "__len__",
        "__setitem__",
        "clear",
        "copy",
        "empty",
        "get",
        "items",
        "keys",
        "pop",
        "popitem",
        "prune",
        "set",
        "setdefault",
        "update",
        "values",
    ),
)
BaseDictProxy._method_to_typeid_ = {
    "__iter__": "Iterator",
}
SyncManager.register("document", datastores.BaseDocument, BaseDictProxy)



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import glob
import json
import os

from directord import components
from directord import utils


# This component was adapted from here:
# https://github.com/openstack/tripleo-ansible/blob/master/tripleo_ansible/ansible_plugins/modules/container_config_data.py  # noqa


DOCUMENTATION = """
---
component: container_config_data
author:
  - Emilien Macchi <emilien@redhat.com>
short_description: Generates a dictionary which contains all container configs
notes: []
description:
  - This module reads container configs in JSON files and generate a dictionary
    which later will be used to manage the containers.
options:
  config_path:
    description:
      - The path of a directory or a file where the JSON files are.
        This parameter is required.
    required: True
    type: str
  config_pattern:
    description:
      - Search pattern to find JSON files.
    default: '*.json'
    required: False
    type: str
  config_overrides:
    description:
      - Allows to override any container configuration which will take
        precedence over the JSON files.
    default: {}
    required: False
    type: dict
  debug:
    description:
      - Whether or not debug is enabled.
    default: False
    required: False
    type: bool
"""


class Component(components.ComponentBase):
    def __init__(self):
        super().__init__(desc="Process echo commands")

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.options_converter(documentation=DOCUMENTATION)

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted cacheevict action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        data.update(vars(self.known_args))
        return data

    def client(self, cache, job):
        """Run cache echo command operation.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        # Set parameters
        config_path = job["config_path"]
        config_pattern = job["config_pattern"]
        config_overrides = job["config_overrides"]

        # Generate dict from JSON files that match search pattern
        if os.path.exists(config_path):
            matched_configs = glob.glob(
                os.path.join(config_path, config_pattern)
            )
            config_dict = {}
            for mc in matched_configs:
                name = os.path.splitext(os.path.basename(mc))[0]
                config = json.loads(self._slurp(mc))
                self.log.debug(
                    "Job [ %s ] Config found for %s: %s",
                    job["job_id"],
                    name,
                    config,
                )
                config_dict.update({name: config})

            # Merge the config dict with given overrides
            configs = utils.merge_dict(config_dict, config_overrides)
            self.set_cache(
                cache=cache,
                key="configs",
                value=configs,
                value_update=False,
                tag="component",
            )
        else:
            self.log.debug(
                "Job [ %s ] %s does not exists, skipping step",
                job["job_id"],
                config_path,
            )
            configs = dict()

        return configs, None, True, None

    def _slurp(self, path):
        """Slurps a file and return its content.

        :param path: string
        :returns: string
        """
        if os.path.exists(path):
            f = open(path, "r")
            return f.read()
        else:
            self.log.warn("%s was not found.", path)
            return ""
